# SPDX-License-Identifier: MIT
# Copyright (c) 2016, Elaborated Networks GmbH
# Copyright (c) 2018, Mathias Laurin

"""Declarations for `mbedtls/pk.h`."""


cimport mbedtls._md as _md
cimport mbedtls.mpi as _mpi


cdef extern from "mbedtls/dhm.h" nogil:
    ctypedef struct mbedtls_dhm_context:
        _mpi.mbedtls_mpi P
        _mpi.mbedtls_mpi G
        _mpi.mbedtls_mpi X
        _mpi.mbedtls_mpi GX
        _mpi.mbedtls_mpi GY
        _mpi.mbedtls_mpi K

    void mbedtls_dhm_init(mbedtls_dhm_context *ctx)
    void mbedtls_dhm_free(mbedtls_dhm_context *ctx)

    int mbedtls_dhm_make_params(
        mbedtls_dhm_context *ctx,
        int x_size,
        unsigned char *output, size_t *olen,
        int (*f_rng)(void *, unsigned char *, size_t), void *p_rng)
    int mbedtls_dhm_make_public(
        mbedtls_dhm_context *ctx,
        int x_size,
        unsigned char *output, size_t olen,
        int (*f_rng)(void *, unsigned char *, size_t), void *p_rng)

    int mbedtls_dhm_read_params(
        mbedtls_dhm_context *ctx,
        unsigned char **p,
        const unsigned char *end)
    int mbedtls_dhm_read_public(
        mbedtls_dhm_context *ctx,
        const unsigned char *input, size_t ilen)

    int mbedtls_dhm_calc_secret(
        mbedtls_dhm_context *ctx,
        unsigned char *output, size_t output_size, size_t *olen,
        int (*f_rng)(void *, unsigned char *, size_t), void *p_rng)

    # int mbedtls_dhm_parse_dhm(
    #     mbedtls_dhm_context *dhm,
    #     const unsigned char *dhmin, size_t dhminlen)
    # int mbedtls_dhm_parse_dhmfile(
    #     mbedtls_dhm_context *dhm,
    #     const char *path)


cdef extern from "mbedtls/ecp.h" nogil:
    ctypedef enum mbedtls_ecp_group_id:
        MBEDTLS_ECP_DP_NONE = 0,
        MBEDTLS_ECP_DP_SECP192R1
        MBEDTLS_ECP_DP_SECP224R1
        MBEDTLS_ECP_DP_SECP256R1
        MBEDTLS_ECP_DP_SECP384R1
        MBEDTLS_ECP_DP_SECP521R1
        MBEDTLS_ECP_DP_BP256R1
        MBEDTLS_ECP_DP_BP384R1
        MBEDTLS_ECP_DP_BP512R1
        MBEDTLS_ECP_DP_CURVE25519
        MBEDTLS_ECP_DP_SECP192K1
        MBEDTLS_ECP_DP_SECP224K1
        MBEDTLS_ECP_DP_SECP256K1
        MBEDTLS_ECP_DP_CURVE448

    ctypedef struct mbedtls_ecp_curve_info:
        mbedtls_ecp_group_id grp_id
        int bit_size
        const char *name

    ctypedef struct mbedtls_ecp_point:
        _mpi.mbedtls_mpi X
        _mpi.mbedtls_mpi Y
        _mpi.mbedtls_mpi Z

    ctypedef struct mbedtls_ecp_group:
        mbedtls_ecp_group_id id;
        _mpi.mbedtls_mpi P
        _mpi.mbedtls_mpi A
        _mpi.mbedtls_mpi B
        mbedtls_ecp_point G
        _mpi.mbedtls_mpi N
        size_t pbits
        size_t nbits
        unsigned int h
        int (*modp)(_mpi.mbedtls_mpi *)
        int (*t_pre)(mbedtls_ecp_point *, void *)
        int (*t_post)(mbedtls_ecp_point *, void *)
        void *t_data
        mbedtls_ecp_point *T
        size_t T_size

    ctypedef struct mbedtls_ecp_keypair:
        mbedtls_ecp_group grp
        _mpi.mbedtls_mpi d
        mbedtls_ecp_point Q

    int MBEDTLS_ECP_MAX_BYTES
    int MBEDTLS_ECP_PF_UNCOMPRESSED

    # Free functions
    # --------------
    const mbedtls_ecp_curve_info* mbedtls_ecp_curve_list()
    # mbedtls_ecp_grp_id_list
    # mbedtls_ecp_curve_info_from_grp_id
    # mbedtls_ecp_curve_info_from_tls_id
    # mbedtls_ecp_curve_info_from_name

    # mbedtls_ecp_point
    # -----------------
    void mbedtls_ecp_point_init(mbedtls_ecp_point *pt)
    void mbedtls_ecp_point_free(mbedtls_ecp_point *pt)
    int mbedtls_ecp_copy(
        mbedtls_ecp_point *P,
        const mbedtls_ecp_point *Q)
    # mbedtls_ecp_set_zero
    int mbedtls_ecp_is_zero(mbedtls_ecp_point *pt)
    int mbedtls_ecp_point_cmp(
        const mbedtls_ecp_point *P,
        const mbedtls_ecp_point *Q)
    # mbedtls_ecp_point_read_string

    # mbedtls_ecp_group
    # -----------------
    void mbedtls_ecp_group_init(mbedtls_ecp_group *grp)
    void mbedtls_ecp_group_free(mbedtls_ecp_group *grp)
    int mbedtls_ecp_group_copy(
        mbedtls_ecp_group *dst,
        const mbedtls_ecp_group *src)

    int mbedtls_ecp_point_write_binary(
        const mbedtls_ecp_group *grp,
        const mbedtls_ecp_point *P,
        int format, size_t *olen, unsigned char *buf, size_t buflen)
    int mbedtls_ecp_point_read_binary(
        const mbedtls_ecp_group *grp,
        mbedtls_ecp_point *P,
        const unsigned char *buf, size_t ilen)

    # mbedtls_ecp_tls_read_point
    # mbedtls_ecp_tls_write_point

    int mbedtls_ecp_group_load(
        mbedtls_ecp_group *grp,
        mbedtls_ecp_group_id index)

    # mbedtls_ecp_tls_read_group
    # mbedtls_ecp_tls_write_group
    int mbedtls_ecp_mul(
        mbedtls_ecp_group *grp,
        mbedtls_ecp_point *R,
        const _mpi.mbedtls_mpi *m,
        const mbedtls_ecp_point *P,
        int (*f_rng)(void *, unsigned char *, size_t),
        void *p_rng)
    # mbedtls_ecp_muladd
    # mbedtls_ecp_check_pubkey
    # mbedtls_ecp_check_privkey

    # mbedtls_ecp_keypair
    # -------------------
    void mbedtls_ecp_keypair_init(mbedtls_ecp_keypair *key)
    void mbedtls_ecp_keypair_free(mbedtls_ecp_keypair *key)
    # mbedtls_ecp_gen_keypair_base
    int mbedtls_ecp_gen_keypair(
        mbedtls_ecp_group *grp,
        _mpi.mbedtls_mpi *d,
        mbedtls_ecp_point *Q,
        int (*f_rng)(void *, unsigned char *, size_t),
        void *p_rng)
    # mbedtls_ecp_check_pub_priv
    int mbedtls_ecp_gen_key(
        mbedtls_ecp_group_id grp_id,
        mbedtls_ecp_keypair *key,
        int (*f_rng)(void *, unsigned char *, size_t),
        void *p_rng)


cdef extern from "mbedtls/ecdh.h" nogil:
    ctypedef enum mbedtls_ecdh_side:
        MBEDTLS_ECDH_OURS
        MBEDTLS_ECDH_THEIRS

    ctypedef struct mbedtls_ecdh_context:
        mbedtls_ecp_group grp
        _mpi.mbedtls_mpi d  # private key
        mbedtls_ecp_point Q  # public key
        mbedtls_ecp_point Qp  # peer's public key
        _mpi.mbedtls_mpi z  # shared secret

    # mbedtls_ecp_group
    # -----------------
    int mbedtls_ecdh_gen_public(
        mbedtls_ecp_group *grp,
        _mpi.mbedtls_mpi *d,
        mbedtls_ecp_point *Q,
        int (*f_rng)(void *, unsigned char *, size_t),
        void *p_rng)
    int mbedtls_ecdh_compute_shared(
        mbedtls_ecp_group *grp,
        _mpi.mbedtls_mpi *z,
        const mbedtls_ecp_point *Q,
        const _mpi.mbedtls_mpi *d,
        int (*f_rng)(void *, unsigned char *, size_t),
        void *p_rng)

    # mbedtls_ecdh_context
    # --------------------
    void mbedtls_ecdh_init(mbedtls_ecdh_context *ctx)
    void mbedtls_ecdh_free(mbedtls_ecdh_context *ctx)

    int mbedtls_ecdh_get_params(
        mbedtls_ecdh_context *ctx,
        const mbedtls_ecp_keypair *key,
        mbedtls_ecdh_side side)

    int mbedtls_ecdh_make_params(
        mbedtls_ecdh_context *ctx,
        size_t *olen, unsigned char *buf, size_t blen,
        int (*f_rng)(void *, unsigned char *, size_t), void *p_rng)
    int mbedtls_ecdh_make_public(
        mbedtls_ecdh_context *ctx,
        size_t *olen, unsigned char *buf, size_t blen,
        int (*f_rng)(void *, unsigned char *, size_t), void *p_rng)

    int mbedtls_ecdh_read_params(
        mbedtls_ecdh_context *ctx,
        const unsigned char **buf, const unsigned char *end)
    int mbedtls_ecdh_read_public(
        mbedtls_ecdh_context *ctx,
        const unsigned char *buf, size_t blen)

    int mbedtls_ecdh_calc_secret(
        mbedtls_ecdh_context *ctx,
        size_t *olen, unsigned char *buf, size_t blen,
        int (*f_rng)(void *, unsigned char *, size_t), void *p_rng)


cdef extern from "mbedtls/ecdsa.h" nogil:
    ctypedef struct mbedtls_ecdsa_context:
        mbedtls_ecp_group grp
        _mpi.mbedtls_mpi d
        mbedtls_ecp_point Q

    int MBEDTLS_ECDSA_MAX_LEN

    # mbedtls_ecp_group
    # -----------------
    # mbedtls_ecdsa_sign
    # mbedtls_ecdsa_sign_det
    # mbedtls_ecdsa_verify

    # mbedtls_ecdsa_context
    # ---------------------
    void mbedtls_ecdsa_init(mbedtls_ecdsa_context *ctx)
    void mbedtls_ecdsa_free(mbedtls_ecdsa_context *ctx)

    int mbedtls_ecdsa_from_keypair(
        mbedtls_ecdsa_context *ctx,
        const mbedtls_ecp_keypair *key)

    # mbedtls_ecdsa_write_signature
    # mbedtls_ecdsa_write_signature_det
    # mbedtls_ecdsa_read_signature

    int mbedtls_ecdsa_genkey(
        mbedtls_ecdsa_context *ctx, mbedtls_ecp_group_id gid,
        int (*f_rng)(void *, unsigned char *, size_t), void *p_rng)


cdef extern from "mbedtls/rsa.h" nogil:
    ctypedef struct mbedtls_rsa_context:
        pass

    # mbedtls_rsa_context
    # -------------------
    # mbedtls_rsa_init
    # mbedtls_rsa_set_padding
    int mbedtls_rsa_gen_key(
        mbedtls_rsa_context *ctx,
        int (*f_rng)(void *, unsigned char *, size_t), void *p_rng,
        unsigned int nbits, int exponent)
    int mbedtls_rsa_check_pubkey(const mbedtls_rsa_context *ctx)
    int mbedtls_rsa_check_privkey(const mbedtls_rsa_context *ctx)
    # mbedtls_rsa_check_pub_priv
    # mbedtls_rsa_public
    # mbedtls_rsa_private
    # mbedtls_rsa_pkcs1_encrypt
    # mbedtls_rsa_rsaes_pkcs1_v15_encrypt
    # mbedtls_rsa_rsaes_oaep_encrypt
    # mbedtls_rsa_pkcs1_decrypt
    # mbedtls_rsa_rsaes_pkcs1_v15_decrypt
    # mbedtls_rsa_rsaes_oaep_decrypt
    # mbedtls_rsa_pkcs1_sign
    # mbedtls_rsa_rsassa_pkcs1_v15_sign
    # mbedtls_rsa_rsassa_pss_sign
    # mbedtls_rsa_pkcs1_verify
    # mbedtls_rsa_rsassa_pkcs1_v15_verify
    # mbedtls_rsa_rsassa_pss_verify
    # mbedtls_rsa_rsassa_pss_verify_ext
    # mbedtls_rsa_copy
    # mbedtls_rsa_free


cdef extern from "mbedtls/pk.h" nogil:
    ctypedef enum mbedtls_pk_type_t:
        MBEDTLS_PK_NONE=0
        MBEDTLS_PK_RSA
        MBEDTLS_PK_ECKEY
        MBEDTLS_PK_ECKEY_DH
        MBEDTLS_PK_ECDSA
        MBEDTLS_PK_RSA_ALT
        MBEDTLS_PK_RSASSA_PSS

    ctypedef struct mbedtls_pk_rsassa_pss_options:
        pass

    ctypedef struct mbedtls_pk_info_t:
        pass
    
    ctypedef struct mbedtls_pk_context:
        pass

    mbedtls_rsa_context *mbedtls_pk_rsa(const mbedtls_pk_context pk)
    mbedtls_ecp_keypair *mbedtls_pk_ec(const mbedtls_pk_context pk)
    # RSA-alt function pointer types
    const mbedtls_pk_info_t *mbedtls_pk_info_from_type(
        mbedtls_pk_type_t pk_type)
    void mbedtls_pk_init(mbedtls_pk_context *ctx)
    void mbedtls_pk_free(mbedtls_pk_context *ctx)
    int mbedtls_pk_setup(mbedtls_pk_context *ctx,
                         const mbedtls_pk_info_t *info)

    size_t mbedtls_pk_get_bitlen(const mbedtls_pk_context *ctx)
    size_t mbedtls_pk_get_len(const mbedtls_pk_context *ctx)
    int mbedtls_pk_can_do(const mbedtls_pk_context *ctx,
                          mbedtls_pk_type_t type)

    int mbedtls_pk_verify(
        mbedtls_pk_context *ctx, _md.mbedtls_md_type_t md_alg,
        const unsigned char *hash, size_t hash_len,
        const unsigned char *sig, size_t sig_len)
    # int mbedtls_pk_verify_ext(
    #     mbedtls_pk_type_t type, const void *options,
    #     mbedtls_pk_context *ctx, _md.mbedtls_md_type_t md_alg,
    #     const unsigned char *hash, size_t hash_len,
    #     const unsigned char *sig, size_t sig_len)

    int mbedtls_pk_sign(
        mbedtls_pk_context *ctx, _md.mbedtls_md_type_t md_alg,
        const unsigned char *hash, size_t hash_len,
        unsigned char *sig, size_t *sig_len,
        int (*f_rng)(void *, unsigned char *, size_t), void *p_rng)

    int mbedtls_pk_decrypt(
        mbedtls_pk_context *ctx,
        const unsigned char *input, size_t ilen,
        unsigned char *output, size_t *olen, size_t osize,
        int (*f_rng)(void *, unsigned char *, size_t), void *p_rng)
    int mbedtls_pk_encrypt(
        mbedtls_pk_context *ctx,
        const unsigned char *input, size_t ilen,
        unsigned char *output, size_t *olen, size_t osize,
        int (*f_rng)(void *, unsigned char *, size_t), void *p_rng)

    int mbedtls_pk_check_pair(const mbedtls_pk_context *pub,
                              const mbedtls_pk_context *prv)
    # int mbedtls_pk_debug(const mbedtls_pk_context *ctx,
    #                      mbedtls_pk_debug_item *items)
    const char * mbedtls_pk_get_name(const mbedtls_pk_context *ctx)
    mbedtls_pk_type_t mbedtls_pk_get_type(const mbedtls_pk_context *ctx)

    int mbedtls_pk_parse_key(
        mbedtls_pk_context *ctx,
        const unsigned char *key, size_t keylen,
        const unsigned char *pwd, size_t pwdlen)
    int mbedtls_pk_parse_public_key(
        mbedtls_pk_context *ctx,
        const unsigned char *key, size_t keylen)

    # int mbedtls_pk_parse_keyfile(
    #     mbedtls_pk_context *ctx,
    #     const char *path, const char *password)
    # int mbedtls_pk_parse_public_keyfile(
    #     mbedtls_pk_context *ctx, const char *path)

    int mbedtls_pk_write_key_der(
        mbedtls_pk_context *ctx,
        unsigned char *buf, size_t size)
    int mbedtls_pk_write_pubkey_der(
        mbedtls_pk_context *ctx,
        unsigned char *buf, size_t size)
    int mbedtls_pk_write_pubkey_pem(
        mbedtls_pk_context *ctx,
        unsigned char *buf, size_t size)
    int mbedtls_pk_write_key_pem(
        mbedtls_pk_context *ctx,
        unsigned char *buf, size_t size)


cdef class CipherBase:
    cdef mbedtls_pk_context _ctx


cdef class RSA(CipherBase):
    pass


cdef class ECC(CipherBase):
    cdef curve


cdef class ECPoint:
    cdef mbedtls_ecp_point _ctx


cdef class DHBase:
    cdef mbedtls_dhm_context _ctx


cdef class ECDHBase:
    cdef mbedtls_ecdh_context _ctx
    cdef curve



# SPDX-License-Identifier: MIT
# Copyright (c) 2016, Elaborated Networks GmbH
# Copyright (c) 2019, Mathias Laurin

"""The cipher package provide symmetric encryption and decryption.

The API follows the recommendations from PEP 272 "API for Block
Encryption Algorithms"

"""

from . import AES, ARC4, ARIA, CHACHA20, DES, DES3, Blowfish, Camellia, DES3dbl
from ._cipher import *
from ._cipher import __all__ as _cipher_all

__all__ = _cipher_all + (
    "AES",
    "ARC4",
    "ARIA",
    "Blowfish",
    "Camellia",
    "DES",
    "DES3",
    "DES3dbl",
    "CHACHA20",
)



# SPDX-License-Identifier: MIT
# Copyright (c) 2018, Mathias Laurin

"""This module defines exceptions and errors."""


from libc.stdlib cimport malloc, free

cimport mbedtls.exceptions as _exc


__all__ = ("TLSError",)


class TLSError(Exception):
    """Exception raise by Mbed TLS."""

    def __init__(self, err=None, msg=""):
        super().__init__()
        if err is not None:
            assert err >= 0
        self.err = err
        self._msg = msg

    @property
    def msg(self):
        if self.err is None:
            return self._msg

        # Set buflen to 200 as in `strerror.c`.
        cdef size_t buflen = 200
        cdef char* buffer = <char*>malloc(buflen * sizeof(char))
        if not buffer:
            raise MemoryError()
        try:
            _exc.mbedtls_strerror(self.err, &buffer[0], buflen)
            output = bytes(buffer[:buflen])
            try:
                olen = output.index(b"\0")
            except ValueError:
                olen = buflen
            return output[:olen].decode("ascii")
        finally:
            free(buffer)

    def __str__(self):
        if self.err is None:
            return "%s(%s)" % (type(self).__name__, self.msg)
        else:
            return "%s([0x%04X] %r)" % (self.__class__.__name__,
                                        self.err, self.msg)


cpdef check_error(const int err):
    if err >= 0:
        return err
    raise TLSError(-err)



# SPDX-License-Identifier: MIT
# Copyright (c) 2018, Mathias Laurin

"""Declarations from `mbedtls/error.h`."""


cdef extern from "mbedtls/error.h" nogil:
    void mbedtls_strerror(int errnum, char *buffer, size_t buflen)



import numbers
import pickle
from binascii import hexlify, unhexlify

import pytest

from mbedtls.mpi import MPI


@pytest.mark.parametrize("value", (12, 2**32 - 1, 10**100))
def test_from_int(value):
    mpi = MPI.from_int(value)
    assert mpi == value
    assert value == mpi
    assert mpi == mpi


@pytest.mark.parametrize("repr_", (repr, str), ids=lambda f: f.__name__)
def test_repr(repr_):
    assert isinstance(repr_(MPI(69)), str)


def test_pickle():
    value = MPI(1337)
    assert value == pickle.loads(pickle.dumps(value))


def test_hash():
    assert isinstance(hash(MPI(1337)), int)


def test_is_integral():
    assert isinstance(MPI(42), numbers.Integral)


def test_prime():
    assert MPI.prime(512).is_prime()


def test_add():
    assert MPI(12) + MPI(12) == 24
    assert MPI(12) + 12 == 24
    assert 12 + MPI(12) == 24


def test_sub():
    assert MPI(12) - MPI(5) == 7
    assert MPI(12) - 5 == 7
    assert 12 - MPI(5) == 7


def test_mul():
    assert MPI(12) * MPI(2) == 24
    assert MPI(12) * 2 == 24
    assert 12 * MPI(2) == 24


def test_pow():
    assert MPI(12).__pow__(5, 12**5 + 1) == 248832
    assert pow(MPI(12), 5, 12**5 + 1) == 248832
    assert MPI(12).__pow__(5, 7) == 3
    assert pow(MPI(12), 5, 7) == 3


def test_eq_same_number_is_true():
    assert (MPI(12) == MPI(12)) is True
    assert (MPI(12) == 12) is True
    assert (12 == MPI(12)) is True


def test_eq_different_numbers_is_false():
    assert (MPI(12) == MPI(42)) is False
    assert (MPI(12) == 42) is False
    assert (12 == MPI(42)) is False


def test_neq_same_numbers_is_false():
    assert (MPI(12) != MPI(12)) is False
    assert (MPI(12) != 12) is False
    assert (12 != MPI(12)) is False


def test_neq_different_numbers_is_true():
    assert (MPI(12) != MPI(42)) is True
    assert (MPI(12) != 42) is True
    assert (12 != MPI(42)) is True


def test_lt_larger_number_is_true():
    assert (MPI(12) < MPI(42)) is True
    assert (MPI(12) < 42) is True
    assert (12 < MPI(42)) is True


def test_lt_smaller_number_is_false():
    assert (MPI(42) < MPI(12)) is False
    assert (MPI(42) < 12) is False
    assert (42 < MPI(12)) is False


def test_lt_same_number_is_false():
    assert (MPI(12) < MPI(12)) is False
    assert (MPI(12) < 12) is False
    assert (12 < MPI(12)) is False


def test_gt_larger_number_is_false():
    assert (MPI(12) > MPI(42)) is False
    assert (MPI(12) > 42) is False
    assert (12 > MPI(42)) is False


def test_gt_smaller_number_is_true():
    assert (MPI(42) > MPI(12)) is True
    assert (MPI(42) > 12) is True
    assert (42 > MPI(12)) is True


def test_gt_same_number_is_false():
    assert (MPI(12) > MPI(12)) is False
    assert (MPI(12) > 12) is False
    assert (12 > MPI(12)) is False


def test_le():
    assert (MPI(12) <= MPI(42)) is True
    assert (MPI(12) <= MPI(12)) is True
    assert (MPI(42) <= MPI(12)) is False


def test_ge():
    assert (MPI(42) >= MPI(12)) is True
    assert (MPI(42) >= MPI(42)) is True
    assert (MPI(12) >= MPI(42)) is False


def test_bool():
    assert bool(MPI(0)) is False


def test_float():
    assert float(MPI(12)) == 12.0


def test_rshift():
    assert MPI(12) >> MPI(2) == 3
    assert MPI(12) >> 2 == 3
    with pytest.raises(TypeError):
        assert 12 >> MPI(2) == 3


def test_lshift():
    assert MPI(12) << MPI(2) == 48
    assert MPI(12) << 2 == 48
    with pytest.raises(TypeError):
        assert 12 << MPI(2) == 48


def test_and():
    assert MPI(12) & MPI(12) == 12
    assert MPI(12) & MPI(3) == 0
    assert MPI(15) & MPI(4) == 4
    assert MPI(15) & 4 == 4
    with pytest.raises(TypeError):
        assert 15 & MPI(4) == 4


def test_or():
    assert MPI(12) | MPI(12) == 12
    assert MPI(12) | MPI(3) == 15
    assert MPI(15) | MPI(4) == 15
    assert MPI(15) | 4 == 15
    with pytest.raises(TypeError):
        assert 15 | MPI(4) == 15


def test_xor():
    assert MPI(12) ^ MPI(12) == 0
    assert MPI(12) ^ MPI(3) == 15
    assert MPI(15) ^ MPI(4) == 11
    assert MPI(15) ^ 4 == 11
    with pytest.raises(TypeError):
        assert 15 ^ MPI(4) == 11


def test_floordiv():
    assert MPI(24) // MPI(2) == 12
    assert MPI(24) // 2 == 12
    assert 24 // MPI(2) == 12


def test_mod():
    assert MPI(12) % MPI(10) == 2
    assert MPI(12) % 10 == 2
    assert 12 % MPI(10) == 2


@pytest.mark.parametrize("value", (12, 2**32 - 1, 10**100))
def test_bit_length(value):
    mpi = MPI(value)
    assert mpi == value
    assert mpi.bit_length() == value.bit_length()


def test_from_empty_bytes():
    value = b""
    big = MPI.from_bytes(value, byteorder="big")
    little = MPI.from_bytes(value, byteorder="little")
    assert big == little == 0
    assert big.bit_length() == little.bit_length() == 0


def test_from_bytes():
    value = unhexlify(b"DEADBEEF")
    mpi = MPI.from_bytes(value, byteorder="big")
    assert mpi.to_bytes(4, byteorder="big") == unhexlify(b"DEADBEEF")
    assert mpi.to_bytes(4, byteorder="little") == unhexlify(b"EFBEADDE")
    assert mpi == int(hexlify(value), 16)


def test_to_bytes_overflow():
    value = unhexlify(b"DEEADBEEFF")
    mpi = MPI.from_bytes(value, byteorder="big")
    with pytest.raises(OverflowError):
        mpi.to_bytes(2, byteorder="big")



import sys
from io import StringIO
from pathlib import Path

import pytest

if sys.version_info < (3, 9):
    from readme_renderer.rst import render
else:
    render = None


@pytest.mark.skipif(render is None, reason="html5lib issue #419")
def test_pypi_rendering():
    # Adapted from `https://stackoverflow.com/questions/46766570/`.
    readme = Path(__file__).parent.parent / "README.rst"
    warnings = StringIO()
    with readme.open() as file_:
        html = render(file_.read(), stream=warnings)
        warnings.seek(0)
        assert html is not None, warnings.read()



# SPDX-License-Identifier: MIT
# Copyright (c) 2018, Mathias Laurin

cimport libc.stdio as c_stdio
from libc.stdlib cimport malloc, free

cimport mbedtls._random as _rnd
cimport mbedtls.pk as _pk
cimport mbedtls._tls as _tls
cimport mbedtls.x509 as _x509

import enum
from collections import abc
from itertools import tee

import certifi
import cython

import mbedtls._random as _rnd
import mbedtls._ringbuf as _rb
import mbedtls.exceptions as _exc
import mbedtls.pk as _pk


cdef _rnd.Random __rng = _rnd.default_rng()


cdef class _PSKSToreProxy:
    def __init__(self, psk_store):
        if not isinstance(psk_store, abc.Mapping):
            raise TypeError("Mapping expected but got %r instead" % psk_store)
        self._mapping = psk_store

    def unwrap(self):
        return self._mapping

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, self._mapping)

    def __str__(self):
        return self._mapping.__str__()

    def __getitem__(self, key):
        return self._mapping.__getitem__(key)

    def __iter__(self):
        return self._mapping.__iter__()

    def __len__(self):
        return self._mapping.__len__()


# Python 2.7: `register()` can be used as a decorator from 3.3.
abc.Mapping.register(_PSKSToreProxy)


@cython.boundscheck(False)
cdef void _my_debug(void *ctx, int level,
                    const char *file, int line, const char *str) nogil:
    c_stdio.fprintf(<c_stdio.FILE *> ctx, "%s:%04d: %s", file, line, str)
    c_stdio.fflush(<c_stdio.FILE *> ctx)


def _enable_debug_output(_BaseConfiguration conf):
    _tls.mbedtls_ssl_conf_dbg(&conf._ctx, _my_debug, c_stdio.stdout)


@cython.boundscheck(False)
cdef int buffer_write(void *ctx, const unsigned char *buf, size_t len) nogil:
    """"Write buffer to output buffer."""
    c_buf = <_tls._C_Buffers *> ctx
    if len == 0:
        return _tls.MBEDTLS_ERR_SSL_BAD_INPUT_DATA
    if _rb.c_len(c_buf.out_ctx) == _rb.c_capacity(c_buf.out_ctx):
        return _tls.MBEDTLS_ERR_SSL_WANT_READ
    cdef size_t writelen = min(
        len, _rb.c_capacity(c_buf.out_ctx) - _rb.c_len(c_buf.out_ctx)
    )
    return _rb.c_write(c_buf.out_ctx, buf, writelen)


@cython.boundscheck(False)
cdef int buffer_read(void *ctx, unsigned char *buf, const size_t len) nogil:
    """Read from input buffer."""
    c_buf = <_tls._C_Buffers *> ctx
    if _rb.c_len(c_buf.in_ctx) == 0:
        return _tls.MBEDTLS_ERR_SSL_WANT_READ
    return _rb.c_readinto(c_buf.in_ctx, buf, len)


@cython.boundscheck(False)
cdef int _psk_cb(
    void *parameter,
    _tls.mbedtls_ssl_context *ctx,
    const unsigned char *c_identity,
    size_t c_identity_len
) nogil:
    """Wrapper for the PSK callback."""
    # If a valid PSK identity is found, call `mbedtls_ssl_set_hs_psk()` and
    # return 0. Otherwise, return 1.
    with gil:
        store = <_tls._PSKSToreProxy> parameter
        identity = c_identity[:c_identity_len]
        try:
            psk = store[identity.decode("utf8")]
            _tls.mbedtls_ssl_set_hs_psk(ctx, psk, len(psk))
            return 0
        except Exception:
            return 1


def _set_debug_level(int level):
    """Set debug level for logging."""
    _tls.mbedtls_debug_set_threshold(level)


def __get_ciphersuite_name(ciphersuite_id):
    """Return a string containing the ciphersuite name.

    Args:
        ciphersuite_id: The ID of the ciphersuite.

    """
    return _tls.mbedtls_ssl_get_ciphersuite_name(
        ciphersuite_id).decode("ascii")


def __get_ciphersuite_id(name):
    """Return the ciphersuite name from ID.

    Args:
        name (str): The name of the ciphersuite.

    """
    cdef char[:] c_name = bytearray(name.encode("ascii"))
    return _tls.mbedtls_ssl_get_ciphersuite_id(&c_name[0])


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def ciphers_available():
    """Return the list of ciphersuites supported by the SSL/TLS module.

    See Also:
        - hashlib.algorithms_available
        - hmac.algorithms_available

    """
    cdef const int* ids = _tls.mbedtls_ssl_list_ciphersuites()
    cdef size_t n = 0
    ciphersuites = []
    while ids[n]:
        ciphersuites.append(__get_ciphersuite_name(ids[n]))
        n += 1
    return tuple(ciphersuites)


@enum.unique
class NextProtocol(enum.Enum):
    # PEP 543
    H2 = b'h2'
    H2C = b'h2c'
    HTTP1 = b'http/1.1'
    WEBRTC = b'webrtc'
    C_WEBRTC = b'c-webrtc'
    FTP = b'ftp'
    STUN = b'stun.nat-discovery'
    TURN = b'stun.turn'


class TLSVersion(enum.IntEnum):
    # MBEDTLS_SSL_|MAJOR|MINOR|_VERSION
    # PEP 543
    # SSLv3 is not safe and is disabled by default.
    # SSLv3 = 0x300
    TLSv1 = 0x301
    TLSv1_1 = 0x302
    TLSv1_2 = 0x303
    MINIMUM_SUPPORTED = TLSv1
    MAXIMUM_SUPPORTED = TLSv1_2

    @classmethod
    def from_major_minor(cls, major, minor):
        return cls((major << 8) + minor)

    def major(self):
        return (self >> 8) & 0xFF

    def minor(self):
        return self & 0xFF


class DTLSVersion(enum.IntEnum):
    DTLSv1_0 = 0x302
    DTLSv1_2 = 0x303
    MINIMUM_SUPPORTED = DTLSv1_0
    MAXIMUM_SUPPORTED = DTLSv1_2

    @classmethod
    def from_major_minor(cls, major, minor):
        return cls((major << 8) + minor)

    def major(self):
        return (self >> 8) & 0xFF

    def minor(self):
        return self & 0xFF


class HandshakeStep(enum.Enum):
    HELLO_REQUEST = _tls.MBEDTLS_SSL_HELLO_REQUEST
    CLIENT_HELLO = _tls.MBEDTLS_SSL_CLIENT_HELLO
    SERVER_HELLO = _tls.MBEDTLS_SSL_SERVER_HELLO
    SERVER_CERTIFICATE = _tls.MBEDTLS_SSL_SERVER_CERTIFICATE
    SERVER_KEY_EXCHANGE = _tls.MBEDTLS_SSL_SERVER_KEY_EXCHANGE
    CERTIFICATE_REQUEST = _tls.MBEDTLS_SSL_CERTIFICATE_REQUEST
    SERVER_HELLO_DONE = _tls.MBEDTLS_SSL_SERVER_HELLO_DONE
    CLIENT_CERTIFICATE = _tls.MBEDTLS_SSL_CLIENT_CERTIFICATE
    CLIENT_KEY_EXCHANGE = _tls.MBEDTLS_SSL_CLIENT_KEY_EXCHANGE
    CERTIFICATE_VERIFY = _tls.MBEDTLS_SSL_CERTIFICATE_VERIFY
    CLIENT_CHANGE_CIPHER_SPEC = _tls.MBEDTLS_SSL_CLIENT_CHANGE_CIPHER_SPEC
    CLIENT_FINISHED = _tls.MBEDTLS_SSL_CLIENT_FINISHED
    SERVER_CHANGE_CIPHER_SPEC = _tls.MBEDTLS_SSL_SERVER_CHANGE_CIPHER_SPEC
    SERVER_FINISHED = _tls.MBEDTLS_SSL_SERVER_FINISHED
    FLUSH_BUFFERS = _tls.MBEDTLS_SSL_FLUSH_BUFFERS
    HANDSHAKE_WRAPUP = _tls.MBEDTLS_SSL_HANDSHAKE_WRAPUP
    HANDSHAKE_OVER = _tls.MBEDTLS_SSL_HANDSHAKE_OVER
    SERVER_NEW_SESSION_TICKET = _tls.MBEDTLS_SSL_SERVER_NEW_SESSION_TICKET
    SERVER_HELLO_VERIFY_REQUEST_SENT = _tls.MBEDTLS_SSL_SERVER_HELLO_VERIFY_REQUEST_SENT


PEM_HEADER = "-----BEGIN CERTIFICATE-----"
PEM_FOOTER = "-----END CERTIFICATE-----"


class WantWriteError(_exc.TLSError):
    pass


class WantReadError(_exc.TLSError):
    pass


class RaggedEOF(_exc.TLSError):
    pass


class HelloVerifyRequest(_exc.TLSError):
    pass


class TrustStore(abc.Sequence):
    def __init__(self, db=None):
        if db is None:
            db = []
        self._db = list(db)

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, self._db)

    @classmethod
    def system(cls):
        return cls.from_pem_file(certifi.where())

    @classmethod
    def from_pem_file(cls, path):
        self = cls()
        with open(str(path)) as cacert:
            pem = None
            for line in cacert.readlines():
                if line.startswith(PEM_HEADER):
                    pem = []
                elif line.strip().endswith(PEM_FOOTER):
                    self.add(_x509.CRT.from_PEM("".join(pem)))
                    pem = None
                elif pem is not None:
                    pem.append(line)
        return self

    def __eq__(self, other):
        if type(other) is not type(self):
            return NotImplemented
        return self._db == other._db

    def __bool__(self):
        return bool(self._db)

    def __len__(self):
        return len(self._db)

    def __getitem__(self, index):
        return self._db[index]

    def add(self, _x509.CRT crt):
        if crt in self:
            return
        cdef _x509.CRT c_crt
        if self._db:
            c_crt = self._db[-1]
            c_crt.set_next(crt)
        self._db.append(crt)


class Purpose(enum.IntEnum):
    SERVER_AUTH = _tls.MBEDTLS_SSL_IS_SERVER
    CLIENT_AUTH = _tls.MBEDTLS_SSL_IS_CLIENT


_DEFAULT_VALUE = object()


cdef class _DTLSCookie:
    """DTLS cookie."""

    def __cinit__(self):
        _tls.mbedtls_ssl_cookie_init(&self._ctx)

    def __dealloc__(self):
        _tls.mbedtls_ssl_cookie_free(&self._ctx)

    @property
    def timeout(self):
        return self._ctx.timeout

    @timeout.setter
    def timeout(self, unsigned long timeout):
        _tls.mbedtls_ssl_cookie_set_timeout(&self._ctx, timeout)

    def generate(self):
        """Generate keys."""
        _tls.mbedtls_ssl_cookie_setup(
            &self._ctx, _rnd.mbedtls_ctr_drbg_random, &__rng._ctx)


cdef class _BaseConfiguration:

    """(D)TLS configuration."""

    def __init__(
        self,
        validate_certificates=None,
        certificate_chain=None,
        ciphers=None,
        inner_protocols=None,
        lowest_supported_version=None,
        highest_supported_version=None,
        trust_store=None,
        sni_callback=None,
        pre_shared_key=None,
        pre_shared_key_store=None,
        _transport=None,
    ):
        _exc.check_error(_tls.mbedtls_ssl_config_defaults(
            &self._ctx,
            endpoint=0,  # server / client is not known here...
            transport=_transport,
            preset=_tls.MBEDTLS_SSL_PRESET_DEFAULT))

        self._set_validate_certificates(validate_certificates)
        self._set_certificate_chain(certificate_chain)
        self._set_ciphers(ciphers)
        self._set_inner_protocols(inner_protocols)
        self._set_lowest_supported_version(lowest_supported_version)
        self._set_highest_supported_version(highest_supported_version)
        self._set_trust_store(trust_store)
        self._set_sni_callback(sni_callback)
        self._set_pre_shared_key(pre_shared_key)
        self._set_pre_shared_key_store(pre_shared_key_store)

        # Set random engine.
        _tls.mbedtls_ssl_conf_rng(
            &self._ctx, &_rnd.mbedtls_ctr_drbg_random, &__rng._ctx)

        # Disable renegotiation.
        _tls.mbedtls_ssl_conf_renegotiation(&self._ctx, 0)

    def __cinit__(self):
        _tls.mbedtls_ssl_config_init(&self._ctx)

        cdef Py_ssize_t ciphers_sz = len(ciphers_available()) + 1
        self._ciphers = <int *>malloc(ciphers_sz * sizeof(int))
        if not self._ciphers:
            raise MemoryError()

        cdef Py_ssize_t idx = 0
        for idx in range(ciphers_sz):
            self._ciphers[idx] = 0

        cdef Py_ssize_t protos_sz = len(NextProtocol) + 1
        self._protos = <const char **>malloc(protos_sz * sizeof(char *))
        if not self._protos:
            raise MemoryError()

        for idx in range(protos_sz):
            self._protos[idx] = NULL

    def __dealloc__(self):
        _tls.mbedtls_ssl_config_free(&self._ctx)
        free(self._ciphers)
        free(self._protos)

    cdef _set_validate_certificates(self, validate):
        """Set the certificate verification mode.

        """  # PEP 543
        if validate is None:
            return
        _tls.mbedtls_ssl_conf_authmode(
            &self._ctx,
            _tls.MBEDTLS_SSL_VERIFY_OPTIONAL if validate is False else
            _tls.MBEDTLS_SSL_VERIFY_REQUIRED)

    @property
    def validate_certificates(self):
        return self._ctx.authmode == _tls.MBEDTLS_SSL_VERIFY_REQUIRED

    cdef _set_certificate_chain(self, chain):
        """The certificate, intermediate certificate, and
        the corresponding private key for the leaf certificate.

        Args:
            chain (Tuple[Tuple[Certificate], PrivateKey]):
                The certificate chain.

        """
        # PEP 543
        if not chain:
            return
        self._chain = chain
        certs, pk_key = chain
        if not certs or not pk_key:
            return
        cdef _x509.CRT c_crt, c_crt_next
        for c_crt, c_crt_next in pairwise(certs):
            c_crt.set_next(c_crt_next)
        c_crt = certs[0]
        c_pk_key = <_pk.CipherBase?> pk_key
        _exc.check_error(_tls.mbedtls_ssl_conf_own_cert(
            &self._ctx, &c_crt._ctx, &c_pk_key._ctx))

    @property
    def certificate_chain(self):
        key_cert = self._ctx.key_cert
        if key_cert is NULL:
            return ((), None)
        chain = []
        cdef _x509.mbedtls_x509_crt *c_ctx = key_cert.cert
        while c_ctx is not NULL:
            chain.append(_x509.CRT.from_DER(c_ctx.raw.p[0:c_ctx.raw.len]))
            c_ctx = c_ctx.next
        cdef _pk.mbedtls_pk_context *c_key = key_cert.key
        cdef unsigned char[:] buf = bytearray(_pk.PRV_DER_MAX_BYTES)
        olen = _exc.check_error(
            _pk.mbedtls_pk_write_key_der(c_key, &buf[0], buf.size))
        cls = {
            0: _pk.ECC,
            1: _pk.RSA,
        }[_pk.mbedtls_pk_can_do(c_key, _pk.MBEDTLS_PK_RSA)]
        key = cls.from_DER(buf[buf.size - olen:buf.size])
        return tuple(chain), key

    cdef _set_ciphers(self, ciphers):
        """The available ciphers for the TLS connections.

        Args:
            ciphers (Tuple[Union[CipherSuite, int]]): The ciphers.

        """ # PEP 543
        if ciphers is None:
            return
        if not frozenset(ciphers).issubset(ciphers_available()):
            raise NotImplementedError("unsupported ciphers")
        cdef Py_ssize_t idx = 0
        self._ciphers[idx] = 0
        for idx, cipher in enumerate(ciphers):
            if not isinstance(cipher, int):
                cipher = __get_ciphersuite_id(cipher)
            self._ciphers[idx] = cipher
        self._ciphers[idx + 1] = 0
        _tls.mbedtls_ssl_conf_ciphersuites(&self._ctx, self._ciphers)

    @property
    def ciphers(self):
        ciphers = []
        cdef int cipher_id
        cdef Py_ssize_t idx
        for idx in range(len(ciphers_available())):
            cipher_id = self._ciphers[idx]
            if cipher_id == 0:
                break
            ciphers.append(__get_ciphersuite_name(cipher_id))
        return tuple(ciphers)

    cdef _set_inner_protocols(self, protocols):
        """

        Args:
            protocols ([Tuple[Union[NextProtocol, bytes]]]): Protocols
                that connections created with this configuration should
                advertise as supported during the TLS handshake. These may
                be advertised using either or both of ALPN or NPN. This
                list of protocols should be ordered by preference.

        """
        # PEP 543
        if protocols is None:
            return
        if len(protocols) > len(NextProtocol):
            raise ValueError("invalid protocols")

        cdef Py_ssize_t idx = 0
        for idx, proto in enumerate(protocols):
            if not isinstance(proto, bytes):
                proto = proto.value
            self._protos[idx] = proto
        self._protos[idx + 1] = NULL

        _exc.check_error(_tls.mbedtls_ssl_conf_alpn_protocols(
            &self._ctx, self._protos))

    @property
    def inner_protocols(self):
        protos = []
        cdef const char* proto
        for idx in range(len(NextProtocol)):
            proto = self._protos[idx]
            if proto is NULL:
                break
            protos.append(NextProtocol(proto))
        return tuple(protos)

    cdef _set_lowest_supported_version(self, version):
        """The minimum version of TLS that should be allowed.

        Args:
            version (TLSVersion, or DTLSVersion): The minimum version.

        """  # PEP 543
        if version is None:
            return
        _tls.mbedtls_ssl_conf_min_version(
            &self._ctx,
            version.major(),
            version.minor(),
        )

    @property
    def _lowest_supported_version(self):
        return self._ctx.min_major_ver, self._ctx.min_minor_ver

    cdef _set_highest_supported_version(self, version):
        """The maximum version of TLS that should be allowed.

        Args:
            version (TLSVersion, or DTLSVersion): The maximum version.

        """  # PEP 543
        if version is None:
            return
        _tls.mbedtls_ssl_conf_max_version(
            &self._ctx,
            version.major(),
            version.minor(),
        )

    @property
    def _highest_supported_version(self):
        return self._ctx.max_major_ver, self._ctx.max_minor_ver

    cdef _set_trust_store(self, store):
        """The trust store that connections will use.

        Args:
            store (TrustStore): The trust store.

        """ # PEP 543
        if not store:
            return
        cdef _x509.CRT crt = TrustStore(store)[0]
        mbedtls_ssl_conf_ca_chain(&self._ctx, &crt._ctx, NULL)

    @property
    def trust_store(self):
        store = TrustStore()
        if self._ctx.ca_chain is NULL:
            return store

        cdef _x509.mbedtls_x509_crt *c_ctx = self._ctx.ca_chain
        while c_ctx is not NULL:
            store.add(_x509.CRT.from_DER(c_ctx.raw.p[0:c_ctx.raw.len]))
            c_ctx = c_ctx.next
        return store

    cdef _set_sni_callback(self, callback):
        # PEP 543, optional, server-side only
        if callback is None:
            return
        # mbedtls_ssl_conf_sni
        raise NotImplementedError

    @property
    def sni_callback(self):
        return None

    cdef _set_pre_shared_key(self, psk):
        """Set a pre shared key (PSK) for the client.

        Args:
            psk ([Tuple[unicode, bytes]]): A tuple with the key and the exected
                identity name.

        """
        if psk is None:
            return
        try:
            identity, key = psk
        except ValueError:
            raise TypeError("expected a tuple (name, key)")
        c_identity = identity.encode("utf8")
        _exc.check_error(_tls.mbedtls_ssl_conf_psk(
            &self._ctx,
            key, len(key),
            c_identity, len(c_identity)))

    @property
    def pre_shared_key(self):
        if self._ctx.psk == NULL or self._ctx.psk_identity == NULL:
            return None
        key = self._ctx.psk[:self._ctx.psk_len]
        c_identity = self._ctx.psk_identity[:self._ctx.psk_identity_len]
        identity = c_identity.decode("utf8")
        return (identity, key)

    cdef _set_pre_shared_key_store(self, psk_store):
        # server-side
        if psk_store is None:
            return
        self._store = _PSKSToreProxy(psk_store)  # ownership
        _tls.mbedtls_ssl_conf_psk_cb(&self._ctx, _psk_cb, <void *> self._store)

    @property
    def pre_shared_key_store(self):
        if self._ctx.p_psk == NULL:
            return None
        psk_store = <_tls._PSKSToreProxy> self._ctx.p_psk
        return psk_store.unwrap()

    def update(self, *args):
        raise NotImplementedError


cdef class TLSConfiguration(_BaseConfiguration):
    """TLS configuration."""
    def __init__(
        self,
        validate_certificates=None,
        certificate_chain=None,
        ciphers=None,
        inner_protocols=None,
        lowest_supported_version=None,
        highest_supported_version=None,
        trust_store=None,
        sni_callback=None,
        pre_shared_key=None,
        pre_shared_key_store=None,
    ):
        super().__init__(
            validate_certificates=validate_certificates,
            certificate_chain=certificate_chain,
            ciphers=ciphers,
            inner_protocols=inner_protocols,
            lowest_supported_version=lowest_supported_version,
            highest_supported_version=highest_supported_version,
            trust_store=trust_store,
            sni_callback=sni_callback,
            pre_shared_key=pre_shared_key,
            pre_shared_key_store=pre_shared_key_store,
            _transport=_tls.MBEDTLS_SSL_TRANSPORT_STREAM,
        )

    @property
    def lowest_supported_version(self):
        return TLSVersion.from_major_minor(*self._lowest_supported_version)

    @property
    def highest_supported_version(self):
        return TLSVersion.from_major_minor(*self._highest_supported_version)

    def __repr__(self):
        return ("%s("
                "validate_certificates=%r, "
                "certificate_chain=%r, "
                "ciphers=%r, "
                "inner_protocols=%r, "
                "lowest_supported_version=%s, "
                "highest_supported_version=%s, "
                "trust_store=%r, "
                "sni_callback=%r, "
                "pre_shared_key=%r, "
                "pre_shared_key_store=%r)"
                % (type(self).__name__,
                   self.validate_certificates,
                   self.certificate_chain,
                   self.ciphers,
                   self.inner_protocols,
                   self.lowest_supported_version,
                   self.highest_supported_version,
                   self.trust_store,
                   self.sni_callback,
                   self.pre_shared_key,
                   self.pre_shared_key_store,
                  ))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return all(
            (
                self.validate_certificates == other.validate_certificates,
                self.certificate_chain == other.certificate_chain,
                self.ciphers == other.ciphers,
                self.inner_protocols == other.inner_protocols,
                self.lowest_supported_version == other.lowest_supported_version,
                self.highest_supported_version == other.highest_supported_version,
                self.trust_store == other.trust_store,
                self.sni_callback == other.sni_callback,
                self.pre_shared_key == other.pre_shared_key,
                self.pre_shared_key_store == other.pre_shared_key_store,
            )
        )

    def __reduce__(self):
        return (
            type(self),
            (
                self.validate_certificates,
                self.certificate_chain,
                self.ciphers,
                self.inner_protocols,
                self.lowest_supported_version,
                self.highest_supported_version,
                self.trust_store,
                self.sni_callback,
                self.pre_shared_key,
                self.pre_shared_key_store,
            ),
        )

    def update(
        self,
        validate_certificates=_DEFAULT_VALUE,
        certificate_chain=_DEFAULT_VALUE,
        ciphers=_DEFAULT_VALUE,
        inner_protocols=_DEFAULT_VALUE,
        lowest_supported_version=_DEFAULT_VALUE,
        highest_supported_version=_DEFAULT_VALUE,
        trust_store=_DEFAULT_VALUE,
        sni_callback=_DEFAULT_VALUE,
        pre_shared_key=_DEFAULT_VALUE,
        pre_shared_key_store=_DEFAULT_VALUE,
    ):
        """Create a new ``TLSConfiguration``.

        Override some of the settings on the original configuration
        with the new settings.

        """
        if validate_certificates is _DEFAULT_VALUE:
            validate_certificates = self.validate_certificates

        if certificate_chain is _DEFAULT_VALUE:
            certificate_chain = self.certificate_chain

        if ciphers is _DEFAULT_VALUE:
            ciphers = self.ciphers

        if inner_protocols is _DEFAULT_VALUE:
            inner_protocols = self.inner_protocols

        if lowest_supported_version is _DEFAULT_VALUE:
            lowest_supported_version = self.lowest_supported_version

        if highest_supported_version is _DEFAULT_VALUE:
            highest_supported_version = self.highest_supported_version

        if trust_store is _DEFAULT_VALUE:
            trust_store = self.trust_store

        if sni_callback is _DEFAULT_VALUE:
            sni_callback = self.sni_callback

        if pre_shared_key is _DEFAULT_VALUE:
            pre_shared_key = self.pre_shared_key

        if pre_shared_key_store is _DEFAULT_VALUE:
            pre_shared_key_store = self.pre_shared_key_store

        return self.__class__(
            validate_certificates=validate_certificates,
            certificate_chain=certificate_chain,
            ciphers=ciphers,
            inner_protocols=inner_protocols,
            lowest_supported_version=lowest_supported_version,
            highest_supported_version=highest_supported_version,
            trust_store=trust_store,
            sni_callback=sni_callback,
            pre_shared_key=pre_shared_key,
            pre_shared_key_store=pre_shared_key_store,
        )


cdef class DTLSConfiguration(_BaseConfiguration):
    """DTLS configuration."""
    def __init__(
        self,
        validate_certificates=None,
        certificate_chain=None,
        ciphers=None,
        inner_protocols=None,
        lowest_supported_version=None,
        highest_supported_version=None,
        trust_store=None,
        anti_replay=None,
        # badmac_limit
        handshake_timeout_min=None,
        handshake_timeout_max=None,
        sni_callback=None,
        pre_shared_key=None,
        pre_shared_key_store=None,
    ):
        super().__init__(
            validate_certificates=validate_certificates,
            certificate_chain=certificate_chain,
            ciphers=ciphers,
            inner_protocols=inner_protocols,
            lowest_supported_version=lowest_supported_version,
            highest_supported_version=highest_supported_version,
            trust_store=trust_store,
            sni_callback=sni_callback,
            pre_shared_key=pre_shared_key,
            pre_shared_key_store=pre_shared_key_store,
            _transport=_tls.MBEDTLS_SSL_TRANSPORT_DATAGRAM,
        )
        self._set_anti_replay(anti_replay)
        self._set_handshake_timeout(handshake_timeout_min, handshake_timeout_max)
        # For security reasons, we do not make cookie optional here.
        cdef _tls._DTLSCookie cookie = _tls._DTLSCookie()
        cookie.generate()
        self._set_cookie(cookie)

    @property
    def lowest_supported_version(self):
        return DTLSVersion.from_major_minor(*self._lowest_supported_version)

    @property
    def highest_supported_version(self):
        return DTLSVersion.from_major_minor(*self._highest_supported_version)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return all(
            (
                self.validate_certificates == other.validate_certificates,
                self.certificate_chain == other.certificate_chain,
                self.ciphers == other.ciphers,
                self.inner_protocols == other.inner_protocols,
                self.lowest_supported_version == other.lowest_supported_version,
                self.highest_supported_version == other.highest_supported_version,
                self.trust_store == other.trust_store,
                self.anti_replay == other.anti_replay,
                self.handshake_timeout_min == other.handshake_timeout_min,
                self.handshake_timeout_max == other.handshake_timeout_max,
                self.sni_callback == other.sni_callback,
                self.pre_shared_key == other.pre_shared_key,
                self.pre_shared_key_store == other.pre_shared_key_store,
            )
        )

    def __repr__(self):
        return ("%s("
                "validate_certificates=%r, "
                "certificate_chain=%r, "
                "ciphers=%r, "
                "inner_protocols=%r, "
                "lowest_supported_version=%s, "
                "highest_supported_version=%s, "
                "trust_store=%r, "
                "anti_replay=%r, "
                "handshake_timeout_min=%r, "
                "handshake_timeout_max=%r, "
                "sni_callback=%r, "
                "pre_shared_key=%r, "
                "pre_shared_key_store=%r)"
                % (type(self).__name__,
                   self.validate_certificates,
                   self.certificate_chain,
                   self.ciphers,
                   self.inner_protocols,
                   self.lowest_supported_version,
                   self.highest_supported_version,
                   self.trust_store,
                   self.anti_replay,
                   self.handshake_timeout_min,
                   self.handshake_timeout_max,
                   self.sni_callback,
                   self.pre_shared_key,
                   self.pre_shared_key_store,
                  ))

    def __reduce__(self):
        return (
            type(self),
            (
                self.validate_certificates,
                self.certificate_chain,
                self.ciphers,
                self.inner_protocols,
                self.lowest_supported_version,
                self.highest_supported_version,
                self.trust_store,
                self.anti_replay,
                self.handshake_timeout_min,
                self.handshake_timeout_max,
                self.sni_callback,
                self.pre_shared_key,
                self.pre_shared_key_store,
            ),
        )

    cdef _set_anti_replay(self, anti_replay):
        """Set anti replay."""
        if anti_replay is None:
            return
        _tls.mbedtls_ssl_conf_dtls_anti_replay(
            &self._ctx,
            _tls.MBEDTLS_SSL_ANTI_REPLAY_ENABLED
            if anti_replay else
            _tls.MBEDTLS_SSL_ANTI_REPLAY_DISABLED)

    @property
    def anti_replay(self):
        cdef unsigned int enabled = _tls.MBEDTLS_SSL_ANTI_REPLAY_ENABLED
        return True if self._ctx.anti_replay == enabled else False

    cdef _set_handshake_timeout(self, minimum, maximum):
        """Set DTLS handshake timeout.

        Args:
            minimum (float, optional): minimum timeout in seconds.
            maximum (float, optional): maximum timeout in seconds.

        """
        if minimum is None and maximum is None:
            return

        def validate(extremum, *, default: float) -> float:
            if extremum is None:
                return default
            if extremum < 0.0:
                raise ValueError(extremum)
            return extremum

        _tls.mbedtls_ssl_conf_handshake_timeout(
            &self._ctx,
            int(1000.0 * validate(minimum, default=1.0)),
            int(1000.0 * validate(maximum, default=60.0)),
        )

    @property
    def handshake_timeout_min(self):
        """Min handshake timeout in seconds (default 1.0)."""
        return float(self._ctx.hs_timeout_min) / 1000.0

    @property
    def handshake_timeout_max(self):
        """Max handshake timeout in seconds (default 60.0)."""
        return float(self._ctx.hs_timeout_max) / 1000.0

    cdef _set_cookie(self, _tls._DTLSCookie cookie):
        """Register callbacks for DTLS cookies (server only)."""
        self._cookie = cookie
        if cookie is None:
            _tls.mbedtls_ssl_conf_dtls_cookies(
                &self._ctx,
                NULL,
                NULL,
                NULL,
            )
        else:
            _tls.mbedtls_ssl_conf_dtls_cookies(
                &self._ctx,
                _tls.mbedtls_ssl_cookie_write,
                _tls.mbedtls_ssl_cookie_check,
                &self._cookie._ctx,
            )

    def update(
        self,
        validate_certificates=_DEFAULT_VALUE,
        certificate_chain=_DEFAULT_VALUE,
        ciphers=_DEFAULT_VALUE,
        inner_protocols=_DEFAULT_VALUE,
        lowest_supported_version=_DEFAULT_VALUE,
        highest_supported_version=_DEFAULT_VALUE,
        trust_store=_DEFAULT_VALUE,
        anti_replay=_DEFAULT_VALUE,
        handshake_timeout_min=_DEFAULT_VALUE,
        handshake_timeout_max=_DEFAULT_VALUE,
        sni_callback=_DEFAULT_VALUE,
        pre_shared_key=_DEFAULT_VALUE,
        pre_shared_key_store=_DEFAULT_VALUE,
    ):
        """Create a new ``DTLSConfiguration``.

        Override some of the settings on the original configuration
        with the new settings.

        """
        if validate_certificates is _DEFAULT_VALUE:
            validate_certificates = self.validate_certificates

        if certificate_chain is _DEFAULT_VALUE:
            certificate_chain = self.certificate_chain

        if ciphers is _DEFAULT_VALUE:
            ciphers = self.ciphers

        if inner_protocols is _DEFAULT_VALUE:
            inner_protocols = self.inner_protocols

        if lowest_supported_version is _DEFAULT_VALUE:
            lowest_supported_version = self.lowest_supported_version

        if highest_supported_version is _DEFAULT_VALUE:
            highest_supported_version = self.highest_supported_version

        if trust_store is _DEFAULT_VALUE:
            trust_store = self.trust_store

        if anti_replay is _DEFAULT_VALUE:
            anti_replay = self.anti_replay

        if handshake_timeout_min is _DEFAULT_VALUE:
            handshake_timeout_min = self.handshake_timeout_min

        if handshake_timeout_max is _DEFAULT_VALUE:
            handshake_timeout_max = self.handshake_timeout_max

        if sni_callback is _DEFAULT_VALUE:
            sni_callback = self.sni_callback

        if pre_shared_key is _DEFAULT_VALUE:
            pre_shared_key = self.pre_shared_key

        if pre_shared_key_store is _DEFAULT_VALUE:
            pre_shared_key_store = self.pre_shared_key_store

        return self.__class__(
            validate_certificates=validate_certificates,
            certificate_chain=certificate_chain,
            ciphers=ciphers,
            inner_protocols=inner_protocols,
            lowest_supported_version=lowest_supported_version,
            highest_supported_version=highest_supported_version,
            trust_store=trust_store,
            anti_replay=anti_replay,
            handshake_timeout_min=handshake_timeout_min,
            handshake_timeout_max=handshake_timeout_max,
            sni_callback=sni_callback,
            pre_shared_key=pre_shared_key,
            pre_shared_key_store=pre_shared_key_store,
        )


DEFAULT_CIPHER_LIST = None


cdef class TLSSession:
    def __cinit__(self):
        """Initialize SSL session structure."""
        _tls.mbedtls_ssl_session_init(&self._ctx)

    def __dealloc__(self):
        """Free referenced items in an SSL session."""
        _tls.mbedtls_ssl_session_free(&self._ctx)

    def __getstate__(self):
        raise TypeError(f"cannot pickle {self.__class__.__name__!r} object")

    def __repr__(self):
        return "%s()" % type(self).__name__


cdef class _BaseContext:
    # _pep543._BaseContext
    """Context base class.

    Args:
        configuration (TLSConfiguration): The configuration.

    """
    def __init__(self, _BaseConfiguration configuration not None):
        self._conf = configuration
        _tls.mbedtls_ssl_conf_endpoint(
            &self._conf._ctx,
            {
                Purpose.CLIENT_AUTH: _tls.MBEDTLS_SSL_IS_CLIENT,
                Purpose.SERVER_AUTH: _tls.MBEDTLS_SSL_IS_SERVER,
            }[self._purpose])

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        if not type(self) is type(other):
            return False
        return self.configuration == other.configuration

    @property
    def configuration(self):
        # PEP 543
        return self._conf

    @property
    def _purpose(self) -> Purpose:
        raise NotImplementedError


TLS_BUFFER_CAPACITY = 2 << 14
# 32K (MBEDTLS_SSL_DTLS_MAX_BUFFERING)


cdef class MbedTLSBuffer:
    def __init__(self, _BaseContext context, server_hostname=None):
        self._context = context
        _exc.check_error(_tls.mbedtls_ssl_setup(&self._ctx, &self._context._conf._ctx))
        self._output_buffer = _rb.RingBuffer(TLS_BUFFER_CAPACITY)
        self._input_buffer = _rb.RingBuffer(TLS_BUFFER_CAPACITY)
        self._c_buffers = _tls._C_Buffers(
            &self._output_buffer._ctx,
            &self._input_buffer._ctx
        )
        self._reset()
        _tls.mbedtls_ssl_set_bio(
            &self._ctx,
            &self._c_buffers,
            buffer_write,
            buffer_read,
            NULL
        )
        self._set_hostname(server_hostname)

    def __cinit__(self, _BaseContext context, server_hostname=None):
        """Initialize an `ssl_context`."""
        _tls.mbedtls_ssl_init(&self._ctx)
        _tls.mbedtls_ssl_set_timer_cb(
            &self._ctx,
            &self._timer,
            _tls.mbedtls_timing_set_delay,
            _tls.mbedtls_timing_get_delay)

    def __dealloc__(self):
        """Free and clear the internal structures of ctx."""
        _tls.mbedtls_ssl_free(&self._ctx)

    def __getstate__(self):
        # TODO:
        #   mbedtls_ssl_context_save(...)
        #   mbdetls_ssl_context_load(...)
        raise TypeError(f"cannot pickle {self.__class__.__name__!r} object")

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, self.context)

    @property
    def context(self):
        return self._context

    @property
    def _verified(self):
        return _tls.mbedtls_ssl_get_verify_result(&self._ctx) == 0

    @property
    def _server_hostname(self):
        # Client side
        if self._ctx.hostname is NULL:
            return None
        return (<bytes> self._ctx.hostname).decode("utf8")

    def _set_hostname(self, hostname):
        """Set the hostname to check against the received server."""
        if hostname is None:
            return
        # Note: `ssl_set_hostname()` makes a copy so it is safe
        #       to call with the temporary `hostname_`.
        hostname_ = hostname.encode("utf8")
        cdef const char* c_hostname = hostname_
        _exc.check_error(_tls.mbedtls_ssl_set_hostname(&self._ctx, c_hostname))

    @property
    def _cookieparam(self):
        """Client ID for the HelloVerifyRequest.

        Note:
            Server-side, DTLS only.

        """
        client_id = bytes(self._ctx.cli_id[0:self._ctx.cli_id_len])
        return client_id if client_id else None

    def setcookieparam(self, const unsigned char[:] info not None):
        if info.size == 0:
            info = b"\0"
        _tls.mbedtls_ssl_set_client_transport_id(
            &self._ctx,
            &info[0],
            info.size,
        )

    def _reset(self):
        _exc.check_error(_tls.mbedtls_ssl_session_reset(&self._ctx))

    def shutdown(self):
        # This could also return SSL_WANT_READ / SSL_WANT_WRITE.
        _exc.check_error(_tls.mbedtls_ssl_close_notify(&self._ctx))
        self._reset()

    def _close(self):
        self.shutdown()

    def read(self, amt):
        # PEP 543
        if amt <= 0:
            return b""
        buffer = bytearray(amt)
        view = memoryview(buffer)
        nread = 0
        while nread != amt and not self._input_buffer.empty():
            nread += self.readinto(view[nread:], amt - nread)
        return bytes(buffer[:nread])

    def readinto(self, unsigned char[:] buffer not None, size_t amt):
        if buffer.size == 0:
            return 0
        if amt <= 0:
            return 0
        read = _tls.mbedtls_ssl_read(&self._ctx, &buffer[0], amt)
        if read > 0:
            return read
        elif read == 0:
            raise RaggedEOF()
        elif read == _tls.MBEDTLS_ERR_SSL_WANT_READ:
            raise WantReadError()
        elif read == _tls.MBEDTLS_ERR_SSL_WANT_WRITE:
            raise WantWriteError()
        elif read == _tls.MBEDTLS_ERR_SSL_CLIENT_RECONNECT:
            _exc.check_error(read)
        else:
            self._reset()
            _exc.check_error(read)

    def write(self, const unsigned char[:] buffer not None):
        if buffer.size == 0:
            return 0
        cdef size_t written = 0
        while written != buffer.size:
            ret = _tls.mbedtls_ssl_write(
                &self._ctx, &buffer[written], buffer.size - written)
            if ret >= 0:
                written += ret
            elif ret == _tls.MBEDTLS_ERR_SSL_WANT_READ:
                raise WantReadError()
            elif ret == _tls.MBEDTLS_ERR_SSL_WANT_WRITE:
                raise WantWriteError()
            else:
                self._reset()
                _exc.check_error(ret)
        assert written == len(buffer)
        return len(self._output_buffer)

    def receive_from_network(self, data):
        # PEP 543
        # Append data to input buffer.
        self._input_buffer.write(data, len(data))

    def peek_outgoing(self, amt):
        # PEP 543
        # Read from output buffer.
        if amt == 0:
            return b""
        return self._output_buffer.peek(amt)

    def consume_outgoing(self, amt):
        """Consume `amt` bytes from the output buffer."""
        # PEP 543
        self._output_buffer.consume(amt)

    def getpeercert(self, binary_form=False):
        """Return the peer certificate, or None."""
        c_crt = _tls.mbedtls_ssl_get_peer_cert(&self._ctx)
        if c_crt is NULL:
            return None

        # See `CRT.to_DER()`.
        der = bytes(c_crt.raw.p[0:c_crt.raw.len])
        if binary_form:
            return der
        return _x509.CRT.from_DER(der)

    def selected_npn_protocol(self):
        return None

    def negotiated_protocol(self):
        cdef const char* protocol = _tls.mbedtls_ssl_get_alpn_protocol(
            &self._ctx)
        if protocol is NULL:
            return None
        return protocol.decode("ascii")

    def _cipher_suite(self):
        cdef const char* name = _tls.mbedtls_ssl_get_ciphersuite(&self._ctx)
        if name is NULL:
            return None
        ssl_version = self.negotiated_tls_version()
        secret_bits = None
        return name.decode("ascii"), ssl_version, secret_bits

    def cipher(self):
        cipher = self._cipher_suite()
        if cipher is None:
            return
        return cipher[0]

    @property
    def _handshake_state(self):
        return HandshakeStep(self._ctx.state)

    def do_handshake(self):
        if self._handshake_state is HandshakeStep.HANDSHAKE_OVER:
            raise ValueError("handshake already over")
        self._handle_handshake_response(
            _tls.mbedtls_ssl_handshake_step(&self._ctx)
        )

    def _renegotiate(self):
        """Initialize an SSL renegotiation on the running connection."""
        self._handle_handshake_response(_tls.mbedtls_ssl_renegotiate(&self._ctx))

    def _handle_handshake_response(self, ret):
        if ret == _tls.MBEDTLS_ERR_SSL_WANT_READ:
            raise WantReadError()
        if ret == _tls.MBEDTLS_ERR_SSL_WANT_WRITE:
            raise WantWriteError()
        if ret == _tls.MBEDTLS_ERR_SSL_HELLO_VERIFY_REQUIRED:
            self._reset()
            raise HelloVerifyRequest()
        if ret < 0:
            self._reset()
            _exc.check_error(ret)
        if ret == 0 and self._output_buffer:
            raise WantWriteError
        assert ret == 0

    def _get_channel_binding(self, cb_type="tls-unique"):
        return None

    def negotiated_tls_version(self):
        # Strings from `ssl_tls.c`.
        return {
            "DTLSv1.0": DTLSVersion.DTLSv1_0,
            "DTLSv1.2": DTLSVersion.DTLSv1_2,
            # "SSLv3.0": TLSVersion.SSLv3,
            "TLSv1.0": TLSVersion.TLSv1,
            "TLSv1.1": TLSVersion.TLSv1_1,
            "TLSv1.2": TLSVersion.TLSv1_2,
        }.get(_tls.mbedtls_ssl_get_version(&self._ctx).decode("ascii"))



import datetime as dt
import pickle
import socket
import subprocess
from contextlib import suppress
from pathlib import Path

import pytest

from mbedtls import hashlib
from mbedtls._tls import _DTLSCookie as DTLSCookie
from mbedtls._tls import _PSKSToreProxy as PSKStoreProxy
from mbedtls.pk import RSA
from mbedtls.tls import *
from mbedtls.tls import HandshakeStep, TLSSession
from mbedtls.x509 import CRT, CSR, BasicConstraints


@pytest.fixture(scope="module")
def rootpath():
    return Path(__file__).parent.parent


class TestPickle:
    @pytest.mark.parametrize(
        "obj",
        (
            TLSConfiguration(),
            DTLSConfiguration(),
            ClientContext(TLSConfiguration()),
            ClientContext(DTLSConfiguration()),
            ServerContext(TLSConfiguration()),
            ServerContext(DTLSConfiguration()),
        ),
        ids=type,
    )
    def test_picklable(self, obj):
        assert obj == pickle.loads(pickle.dumps(obj))

    @pytest.mark.parametrize(
        "obj",
        (
            TLSSession(),
            TLSWrappedBuffer(ClientContext(DTLSConfiguration())),
            TLSWrappedBuffer(ClientContext(TLSConfiguration())),
            TLSWrappedBuffer(ServerContext(DTLSConfiguration())),
            TLSWrappedBuffer(ServerContext(TLSConfiguration())),
            TLSWrappedSocket(
                socket.socket(),
                TLSWrappedBuffer(ClientContext(DTLSConfiguration())),
            ),
            TLSWrappedSocket(
                socket.socket(),
                TLSWrappedBuffer(ClientContext(TLSConfiguration())),
            ),
            TLSWrappedSocket(
                socket.socket(),
                TLSWrappedBuffer(ServerContext(DTLSConfiguration())),
            ),
            TLSWrappedSocket(
                socket.socket(),
                TLSWrappedBuffer(ServerContext(TLSConfiguration())),
            ),
        ),
        ids=type,
    )
    def test_unpicklable(self, obj):
        with pytest.raises(TypeError) as excinfo:
            pickle.dumps(obj)

        assert str(excinfo.value).startswith("cannot pickle")


class TestPSKStoreProxy:
    @pytest.fixture
    def psk_store(self):
        return {"client": b"the secret key"}

    @pytest.fixture
    def proxy(self, psk_store):
        return PSKStoreProxy(psk_store)

    @pytest.mark.parametrize("repr_", (repr, str), ids=lambda f: f.__name__)
    def test_repr(self, repr_, psk_store):
        assert isinstance(repr_(psk_store), str)

    def test_unwrap(self, proxy, psk_store):
        assert proxy.unwrap() == psk_store

    def test_eq(self, proxy, psk_store):
        for k, v in psk_store.items():
            assert proxy[k] == v

    def test_len(self, proxy, psk_store):
        assert len(proxy) == len(psk_store)


class TestTLSVersion:
    @pytest.mark.parametrize("version", TLSVersion)
    def test_major(self, version):
        assert version.major() == 3

    def test_minor(self):
        # assert TLSVersion.SSLv3.minor() == 0
        assert TLSVersion.TLSv1.minor() == 1
        assert TLSVersion.TLSv1_1.minor() == 2
        assert TLSVersion.TLSv1_2.minor() == 3

    @pytest.mark.parametrize("version", TLSVersion)
    def test_from_major_minor(self, version):
        assert (
            TLSVersion.from_major_minor(version.major(), version.minor())
            is version
        )

    @pytest.mark.parametrize(
        "version", [TLSVersion.MINIMUM_SUPPORTED, TLSVersion.MAXIMUM_SUPPORTED]
    )
    def test_minmax_supported(self, version):
        assert version in TLSVersion


class TestDTLSVersion:
    @pytest.mark.parametrize("version", DTLSVersion)
    def test_major(self, version):
        assert version.major() == 3

    def test_minor(self):
        assert DTLSVersion.DTLSv1_0.minor() == 2
        assert DTLSVersion.DTLSv1_2.minor() == 3

    @pytest.mark.parametrize("version", DTLSVersion)
    def test_from_major_minor(self, version):
        assert (
            DTLSVersion.from_major_minor(version.major(), version.minor())
            is version
        )

    @pytest.mark.parametrize(
        "version",
        [DTLSVersion.MINIMUM_SUPPORTED, DTLSVersion.MAXIMUM_SUPPORTED],
    )
    def test_minmax_supported(self, version):
        assert version in DTLSVersion


class TestTLSRecordHeader:
    @pytest.fixture(params=TLSRecordHeader.RecordType)
    def record_type(self, request):
        return request.param

    @pytest.fixture(params=TLSVersion)
    def version(self, request):
        return request.param

    @pytest.fixture
    def length(self):
        return 42

    @pytest.fixture
    def header(self, record_type, version, length):
        return TLSRecordHeader(record_type, version, length)

    @pytest.mark.parametrize("repr_", (repr, str), ids=lambda f: f.__name__)
    def test_repr(self, repr_, record_type):
        assert isinstance(repr_(record_type), str)

    def test_hash(self, record_type):
        assert isinstance(hash(record_type), int)

    def test_accessors(self, header, record_type, version, length):
        assert len(header) == 5
        assert header.record_type is record_type
        assert header.version is version
        assert header.length == length

    def test_serialization(self, header):
        serialized = bytes(header)
        assert isinstance(serialized, bytes)
        assert len(serialized) == 5
        assert TLSRecordHeader.from_bytes(serialized) == header


class TestTLSSession:
    @pytest.fixture
    def session(self):
        return TLSSession()

    def test_repr(self, session):
        assert isinstance(repr(session), str)


class Chain:
    @pytest.fixture(scope="class")
    def now(self):
        return dt.datetime.utcnow()

    @pytest.fixture(scope="class")
    def digestmod(self):
        return hashlib.sha256

    @pytest.fixture(scope="class")
    def ca0_key(self):
        ca0_key = RSA()
        ca0_key.generate()
        return ca0_key

    @pytest.fixture(scope="class")
    def ca1_key(self):
        ca1_key = RSA()
        ca1_key.generate()
        return ca1_key

    @pytest.fixture(scope="class")
    def ee0_key(self):
        ee0_key = RSA()
        ee0_key.generate()
        return ee0_key

    @pytest.fixture(scope="class")
    def ca0_crt(self, ca0_key, digestmod, now):
        ca0_csr = CSR.new(ca0_key, "CN=Trusted CA", digestmod())
        return CRT.selfsign(
            ca0_csr,
            ca0_key,
            not_before=now,
            not_after=now + dt.timedelta(days=90),
            serial_number=0x123456,
            basic_constraints=BasicConstraints(True, -1),
        )

    @pytest.fixture(scope="class")
    def ca1_crt(self, ca1_key, ca0_crt, ca0_key, digestmod, now):
        ca1_csr = CSR.new(ca1_key, "CN=Intermediate CA", digestmod())
        return ca0_crt.sign(
            ca1_csr,
            ca0_key,
            now,
            now + dt.timedelta(days=90),
            0x234567,
            basic_constraints=BasicConstraints(True, -1),
        )

    @pytest.fixture(scope="class")
    def ee0_crt(self, ee0_key, ca1_crt, ca1_key, digestmod, now):
        ee0_csr = CSR.new(ee0_key, "CN=End Entity", digestmod())
        return ca1_crt.sign(
            ee0_csr, ca1_key, now, now + dt.timedelta(days=90), 0x345678
        )

    @pytest.fixture(scope="class")
    def certificate_chain(self, ee0_crt, ca1_crt, ee0_key):
        return (ee0_crt, ca1_crt), ee0_key


class TestTrustStore(Chain):
    @pytest.fixture
    def store(self):
        return TrustStore.system()

    @pytest.mark.parametrize("repr_", (repr, str), ids=lambda f: f.__name__)
    def test_repr(self, repr_, store):
        assert isinstance(repr_(store), str)

    def test_eq(self, store):
        other = TrustStore(store)
        assert store is not other
        assert store == other

    def test_bool(self, store):
        assert not TrustStore()
        assert store

    def test_len(self, store):
        assert len(store) != 0

    def test_iter(self, store):
        assert store[0] != store[1]
        for n, crt in enumerate(store, start=1):
            assert crt in store
        assert n == len(store)

    def test_add_existing_certificate(self, store):
        length = len(store)
        store.add(store[0])
        assert len(store) == length

    def test_add_new_certificate(self, store, ca0_crt):
        length = len(store)
        store.add(ca0_crt)
        assert len(store) == length + 1


class TestDTLSCookie:
    @pytest.fixture
    def cookie(self):
        return DTLSCookie()

    def test_generate_does_not_raise(self, cookie):
        cookie.generate()

    def test_timeout(self, cookie):
        assert cookie.timeout == 60
        cookie.timeout = 1000
        assert cookie.timeout == 1000


class _BaseConfiguration(Chain):
    @pytest.fixture
    def conf(self):
        raise NotImplementedError

    @pytest.fixture
    def version(self):
        raise NotImplementedError

    @pytest.mark.parametrize("repr_", (repr, str), ids=lambda f: f.__name__)
    def test_repr(self, repr_, conf):
        assert isinstance(repr_(conf), str)

    @pytest.mark.parametrize("validate", [True, False])
    def test_set_validate_certificates(self, conf, validate):
        conf_ = conf.update(validate_certificates=validate)
        assert conf_.validate_certificates is validate

    @pytest.mark.parametrize("chain", [((), None), None])
    def test_set_certificate_chain(self, conf, chain, certificate_chain):
        if chain is None:
            chain = certificate_chain
        conf_ = conf.update(certificate_chain=chain)
        assert conf_.certificate_chain == chain

    @pytest.mark.parametrize("ciphers", (ciphers_available(),))
    def test_set_ciphers(self, conf, ciphers):
        conf_ = conf.update(ciphers=ciphers)
        assert conf_.ciphers == ciphers

    @pytest.mark.parametrize(
        "inner_protocols",
        [[], (), [NextProtocol.H2, NextProtocol.H2C], [b"h2", b"h2c", b"ftp"]],
    )
    def test_set_inner_protocols(self, conf, inner_protocols):
        conf_ = conf.update(inner_protocols=inner_protocols)
        assert conf_.inner_protocols == tuple(
            NextProtocol(_) for _ in inner_protocols
        )

    def test_lowest_supported_version(self, conf, version):
        conf_ = conf.update(lowest_supported_version=version)
        assert conf_.lowest_supported_version is version

    def test_highest_supported_version(self, conf, version):
        conf_ = conf.update(highest_supported_version=version)
        assert conf_.highest_supported_version is version

    @pytest.mark.parametrize("store", [TrustStore.system()])
    def test_trust_store(self, conf, store):
        conf_ = conf.update(trust_store=store)
        assert store
        assert conf_.trust_store == store

    @pytest.mark.parametrize("callback", [None])
    def test_set_sni_callback(self, conf, callback):
        assert conf.sni_callback is None

    @pytest.mark.parametrize("psk", [None, ("client", b"the secret key")])
    def test_psk(self, conf, psk):
        assert conf.pre_shared_key is None
        conf_ = conf.update(pre_shared_key=psk)
        assert conf_.pre_shared_key == psk

    @pytest.mark.parametrize(
        "psk_store", [None, {"client": b"the secret key"}]
    )
    def test_psk_store(self, conf, psk_store):
        assert conf.pre_shared_key_store is None
        conf_ = conf.update(pre_shared_key_store=psk_store)
        assert conf_.pre_shared_key_store == psk_store


class TestTLSConfiguration(_BaseConfiguration):
    @pytest.fixture
    def conf(self):
        return TLSConfiguration()

    @pytest.fixture(params=TLSVersion)
    def version(self, request):
        return request.param


class TestDTLSConfiguration(_BaseConfiguration):
    @pytest.fixture
    def conf(self):
        return DTLSConfiguration()

    @pytest.fixture(params=DTLSVersion)
    def version(self, request):
        return request.param

    @pytest.mark.parametrize("anti_replay", [True, False])
    def test_set_anti_replay(self, conf, anti_replay):
        assert conf.anti_replay is True
        conf_ = conf.update(anti_replay=anti_replay)
        assert conf_.anti_replay is anti_replay

    @pytest.mark.parametrize(
        "hs_min, hs_max", [(1, 60), (42, 69), (4.2, 6.9), (42.0, 69.0)]
    )
    def test_handshake_timeout_minmax(self, conf, hs_min, hs_max):
        assert conf.handshake_timeout_min == 1.0
        assert conf.handshake_timeout_max == 60.0
        conf_ = conf.update(
            handshake_timeout_min=hs_min,
            handshake_timeout_max=hs_max,
        )
        assert conf_.handshake_timeout_min == hs_min
        assert conf_.handshake_timeout_max == hs_max

    @pytest.mark.parametrize(
        "hs_min, hs_max", [(None, None), (1, None), (None, 60)]
    )
    def test_handshake_timeout_default(self, conf, hs_min, hs_max):
        conf_ = conf.update(
            handshake_timeout_min=hs_min,
            handshake_timeout_max=hs_max,
        )
        assert conf_.handshake_timeout_min == hs_min or 1.0
        assert conf_.handshake_timeout_max == hs_max or 60.0


class TestBaseContext:
    @pytest.fixture(params=[Purpose.SERVER_AUTH, Purpose.CLIENT_AUTH])
    def purpose(self, request):
        return request.param

    @pytest.fixture(params=[TLSConfiguration, DTLSConfiguration])
    def conf(self, request):
        return request.param()

    @pytest.fixture(params=[ServerContext, ClientContext])
    def context(self, conf, request):
        cls = request.param
        return cls(conf)

    def test_repr(self, context):
        assert isinstance(repr(context), str)

    def test_get_configuration(self, context, conf):
        assert conf
        assert context.configuration is conf


class TestClientContext(TestBaseContext):
    @pytest.fixture(params=[None, "hostname", "localhost"])
    def hostname(self, request):
        return request.param

    @pytest.fixture
    def context(self, conf, hostname):
        return ClientContext(conf)

    def test_context(self, context):
        assert isinstance(context, ClientContext)

    def test_hostname(self, context, hostname):
        wrapped_buffers = context.wrap_buffers(hostname)
        assert wrapped_buffers._server_hostname == hostname

    def test_wrap_buffers(self, context):
        assert isinstance(context.wrap_buffers(None), TLSWrappedBuffer)


class TestServerContext(TestBaseContext):
    @pytest.fixture
    def context(self, conf):
        return ServerContext(conf)

    def test_context(self, context):
        assert isinstance(context, ServerContext)

    def test_wrap_buffers(self, context):
        assert isinstance(context.wrap_buffers(), TLSWrappedBuffer)


CLIENT_HELLO = (HandshakeStep.CLIENT_HELLO,)
SERVER_HELLO = (
    HandshakeStep.SERVER_HELLO,
    HandshakeStep.SERVER_CERTIFICATE,
    HandshakeStep.SERVER_KEY_EXCHANGE,
    HandshakeStep.CERTIFICATE_REQUEST,
    HandshakeStep.SERVER_HELLO_DONE,
)
CLIENT_KEY_EXCHANGE = (
    HandshakeStep.CLIENT_KEY_EXCHANGE,
    HandshakeStep.CERTIFICATE_VERIFY,
    HandshakeStep.CLIENT_CHANGE_CIPHER_SPEC,
    HandshakeStep.CLIENT_FINISHED,
)
CHANGE_CIPHER_SPEC = (
    HandshakeStep.SERVER_CHANGE_CIPHER_SPEC,
    HandshakeStep.SERVER_FINISHED,
    HandshakeStep.FLUSH_BUFFERS,
    HandshakeStep.HANDSHAKE_WRAPUP,
    HandshakeStep.HANDSHAKE_OVER,
)


def do_io(*, src, dst, amt=1024):
    __tracebackhide__ = True
    in_transit = src.peek_outgoing(amt)
    src.consume_outgoing(len(in_transit))
    dst.receive_from_network(in_transit)


def do_handshake(end, states):
    __tracebackhide__ = True
    while end._handshake_state is not states[0]:
        # The backend goes through every state for both
        # ends.  This is not relevant.
        end.do_handshake()

    for state in states:
        assert end._handshake_state is state
        if state is HandshakeStep.HANDSHAKE_OVER:
            break
        with suppress(WantReadError, WantWriteError):
            end.do_handshake()


def make_full_handshake(*, client, server):
    do_handshake(client, CLIENT_HELLO)
    do_io(src=client, dst=server)
    do_handshake(server, SERVER_HELLO)
    assert client.negotiated_protocol() == server.negotiated_protocol()

    do_io(src=server, dst=client)
    do_handshake(client, CLIENT_KEY_EXCHANGE)
    assert client.negotiated_tls_version() == server.negotiated_tls_version()

    do_io(src=client, dst=server)
    do_handshake(server, CHANGE_CIPHER_SPEC)
    do_io(src=server, dst=client)
    do_handshake(client, (HandshakeStep.HANDSHAKE_OVER,))
    assert client.cipher() == server.cipher()


def make_hello_verify_request(*, client, server, cookie):
    do_handshake(client, CLIENT_HELLO)
    do_io(src=client, dst=server)
    server.setcookieparam(cookie)
    with pytest.raises(HelloVerifyRequest):
        do_handshake(
            server,
            (
                HandshakeStep.SERVER_HELLO,
                HandshakeStep.SERVER_HELLO_VERIFY_REQUEST_SENT,
            ),
        )

    do_handshake(server, (HandshakeStep.HELLO_REQUEST,))
    server.setcookieparam(cookie)
    do_io(src=server, dst=client)


class TestHandshake:
    def test_tls(self):
        psk = ("cli", b"secret")

        srv_ctx = ServerContext(
            TLSConfiguration(
                pre_shared_key_store=dict((psk,)),
                validate_certificates=False,
            )
        )
        cli_ctx = ClientContext(
            TLSConfiguration(
                pre_shared_key=psk,
                validate_certificates=False,
            )
        )
        server = srv_ctx.wrap_buffers()
        client = cli_ctx.wrap_buffers("hostname")

        make_full_handshake(client=client, server=server)

        secret = b"a very secret message"

        amt = client.write(secret)
        do_io(src=client, dst=server)
        assert server.read(amt) == secret

        amt = server.write(secret)
        do_io(src=server, dst=client)
        assert client.read(amt) == secret

    def test_dtls(self):
        psk = ("cli", b"secret")

        srv_ctx = ServerContext(
            DTLSConfiguration(
                pre_shared_key_store=dict((psk,)),
                validate_certificates=False,
            )
        )
        cli_ctx = ClientContext(
            DTLSConfiguration(
                pre_shared_key=psk,
                validate_certificates=False,
            )
        )
        server = srv_ctx.wrap_buffers()
        client = cli_ctx.wrap_buffers("hostname")

        make_hello_verify_request(
            client=client, server=server, cookie="🍪🍪🍪".encode("utf-8")
        )
        make_full_handshake(client=client, server=server)

        secret = b"a very secret message"

        amt = client.write(secret)
        do_io(src=client, dst=server)
        assert server.read(amt) == secret

        amt = server.write(secret)
        do_io(src=server, dst=client)
        assert client.read(amt) == secret


@pytest.mark.local
class TestProgramsTLS:
    @pytest.fixture
    def port(self):
        """Return a free port

        Note:
            Not 100% race condition free.

        """
        port = 0
        with socket.socket() as sock:
            sock.bind(("", port))
            port = sock.getsockname()[1]
        return port

    @pytest.fixture
    def server(self, rootpath, port):
        args = [
            rootpath / "programs" / "server.py",
            "--port",
            f"{port}",
            "--tls",
            "--psk-store",
            "cli=secret",
        ]
        proc = subprocess.Popen(args)
        yield proc
        proc.kill()
        proc.wait(1.0)

    @pytest.mark.usefixtures("server")
    @pytest.mark.timeout(10)
    def test_communicate(self, rootpath, port):
        secret = b"a very secret message"
        args = [
            rootpath / "programs" / "client.py",
            "--port",
            f"{port}",
            "--tls",
            "--psk",
            "cli=secret",
            secret,
        ]
        for _ in range(3):
            with subprocess.Popen(args, stdout=subprocess.PIPE) as client:
                out, err = client.communicate()
                assert out == secret + b"\n"


@pytest.mark.local
class TestProgramsDTLS:
    @pytest.fixture
    def port(self):
        """Return a free port

        Note:
            Not 100% race condition free.

        """
        port = 0
        with socket.socket() as sock:
            sock.bind(("", port))
            port = sock.getsockname()[1]
        return port

    @pytest.fixture
    def server(self, rootpath, port):
        args = [
            rootpath / "programs" / "server.py",
            "--port",
            f"{port}",
            "--dtls",
            "--psk-store",
            "cli=secret",
        ]
        proc = subprocess.Popen(args)
        yield proc
        proc.kill()
        proc.wait(1.0)

    @pytest.mark.usefixtures("server")
    @pytest.mark.timeout(10)
    def test_communication(self, rootpath, port):
        secret = b"a very secret message"
        args = [
            rootpath / "programs" / "client.py",
            "--port",
            f"{port}",
            "--dtls",
            "--psk",
            "cli=secret",
            secret,
        ]
        for _ in range(3):
            with subprocess.Popen(args, stdout=subprocess.PIPE) as client:
                out, err = client.communicate()
                assert out == secret + b"\n"



# SPDX-License-Identifier: MIT
# Copyright (c) 2019, Mathias Laurin

"""Declarations from `mbedtls/platform*.h`."""


cdef extern from "mbedtls/platform_util.h" nogil:
    void mbedtls_platform_zeroize(void *buf, size_t len)



# SPDX-License-Identifier: MIT
# Copyright (c) 2015, Elaborated Networks GmbH
# Copyright (c) 2019, Mathias Laurin

"""Generic message digest wrapper (hash algorithm)."""


import mbedtls._md as _md

algorithms_guaranteed = _md.algorithms_guaranteed
algorithms_available = _md.algorithms_available


Hmac = _md.Hmac


def new(key, buffer=None, digestmod=None):
    """A generic constructor that takes the key algorithm as its first
    parameter.

    """
    if digestmod is None:
        digestmod = "md5"
    return Hmac(key, digestmod, buffer)


def md2(key, buffer=None):
    """MD2 message-digest algorithm."""
    return Hmac(key, "md2", buffer)


def md4(key, buffer=None):
    """MD4 message-digest algorithm."""
    return Hmac(key, "md4", buffer)


def md5(key, buffer=None):
    """MD5 message-digest algorithm."""
    return Hmac(key, "md5", buffer)


def sha1(key, buffer=None):
    """Secure Hmac Algorithm 1 (SHA-1)."""
    return Hmac(key, "sha1", buffer)


def sha224(key, buffer=None):
    """Secure Hmac Algorithm 2 (SHA-2) with 224 bits hash value."""
    return Hmac(key, "sha224", buffer)


def sha256(key, buffer=None):
    """Secure Hmac Algorithm 2 (SHA-2) with 256 bits hash value."""
    return Hmac(key, "sha256", buffer)


def sha384(key, buffer=None):
    """Secure Hmac Algorithm 2 (SHA-2) with 384 bits hash value."""
    return Hmac(key, "sha384", buffer)


def sha512(key, buffer=None):
    """Secure Hmac Algorithm 2 (SHA-2) with 512 bits hash value."""
    return Hmac(key, "sha512", buffer)


def ripemd160(key, buffer=None):
    """RACE Integrity Primitives Evaluation Message Digest (RIPEMD) with
    160 bits hash value."""
    return Hmac(key, "ripemd160", buffer)



import random
import sys

import pytest

import mbedtls

try:
    from collections.abc import Sequence
except ImportError:
    from collections import Sequence
try:
    import reprlib
except ImportError:
    import repr as reprlib


def pytest_report_header(config, startdir):
    sys.stdout.write(
        "python-mbedtls {0}, {1}\n".format(
            mbedtls.__version__, mbedtls.version.version
        )
    )


class _Repr(reprlib.Repr):
    """Repr with support for memoryview."""

    def repr_memoryview(self, obj, level):
        return "%s(%s)" % (type(obj).__name__, self.repr(obj.tobytes()))


_repr_instance = _Repr()
_repr = _repr_instance.repr


def issequence(x):
    # Adapted from pytest.
    if bytes != str:
        return isinstance(x, Sequence) and not isinstance(x, str)
    else:
        return isinstance(x, Sequence)


def _compare_memoryviews(_config, op, left, right):
    # Adapted from pytest.
    summary = ["{} != {}".format(_repr(left), _repr(right))]
    explanation = []
    if issequence(left) and issequence(right):
        for i in range(min(len(left), len(right))):
            if left[i] != right[i]:
                left_value = left[i : i + 1]
                right_value = right[i : i + 1]
                explanation += [
                    "At index {} diff: {} != {}".format(
                        i, _repr(left_value), _repr(right_value)
                    )
                ]
                break
    return summary + explanation


def pytest_assertrepr_compare(config, op, left, right):
    if op == "==" and any(
        (isinstance(left, memoryview), isinstance(right, memoryview))
    ):
        return _compare_memoryviews(config, op, left, right)
    return None


@pytest.fixture
def randbytes():
    def function(length):
        return bytes(
            bytearray(random.randrange(0, 256) for _ in range(length))
        )

    return function


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # set a report attribute for each phase of a call, which can
    # be "setup", "call", "teardown"

    setattr(item, "rep_" + rep.when, rep)



# SPDX-License-Identifier: MIT
# Copyright (c) 2018, Mathias Laurin

"""Structure and functions for parsing and writing X.509 certificates."""


from libc.stdlib cimport malloc, free

cimport mbedtls.x509 as x509
cimport mbedtls.mpi as _mpi
cimport mbedtls.pk as _pk

import base64
import datetime as dt
from collections import namedtuple
from contextlib import suppress

import mbedtls.exceptions as _exc
import mbedtls.mpi as _mpi
import mbedtls.pk as _pk
import mbedtls._md as _md
from mbedtls.hashlib import new as _new_hash

import enum


@enum.unique
class KeyUsage(enum.IntEnum):
    """Key Usage Extension.

    See Also:
        RFC 5280 - 4.2.1.3 Key Usage.

    """
    DIGITAL_SIGNATURE  = 0x80
    NON_REPUDIATION = 0x40
    KEY_ENCIPHERMENT  = 0x20
    DATA_ENCIPHERMENT = 0x10
    KEY_AGREEMENT = 0x08
    KEY_CERT_SIGN = 0x04
    CRL_SIGN = 0x02
    ENCIPHER_ONLY = 0x01
    DECIPHER_ONLY = 0x8000


def PEM_to_DER(pem):
    return base64.b64decode(
        b"".join(line.encode("ascii") for line in pem.splitlines()
                 if not line.startswith("-----")))


def DER_to_PEM(der, text):
    chunk_size = 64
    pem = base64.b64encode(der).decode("ascii")
    return "\n".join((
        "-----BEGIN %s-----" % text.upper(),
        "\n".join(pem[n:n+chunk_size] for n in range(0, len(pem), chunk_size)),
        "-----END %s-----" % text.upper(),
        ""))


cdef class Certificate:
    @classmethod
    def from_buffer(cls, buffer):
        # PEP 543
        return cls(buffer)

    @classmethod
    def from_file(cls, path):
        # PEP 543
        raise NotImplementedError

    @classmethod
    def from_DER(cls, der):
        raise NotImplementedError

    @classmethod
    def from_PEM(cls, pem):
        raise NotImplementedError

    def __reduce__(self):
        return type(self).from_DER, (bytes(self),)

    def __hash__(self):
        return hash(self.to_DER())

    def __eq__(self, other):
        if type(other) is type(self):
            return self.to_DER() == other.to_DER()
        else:
            return other == self.to_DER() or other == self.to_PEM()

    def __str__(self):
        raise NotImplementedError

    def __bytes__(self):
        return self.to_DER()

    def export(self, format="DER"):
        if format == "DER":
            return self.to_DER()
        if format == "PEM":
            return self.to_PEM()
        raise ValueError(format)

    def to_bytes(self):
        return self.to_DER()

    def to_DER(self):
        raise NotImplementedError

    def to_PEM(self):
        raise NotImplementedError


class BasicConstraints(
        namedtuple("BasicConstraints", ["ca", "max_path_length"])):
    """The basic constraints for the certificate."""

    def __new__(cls, ca=False, max_path_length=0):
        return super().__new__(cls, ca, max_path_length)


cdef class CRT(Certificate):
    """X.509 certificate."""

    def __init__(self, const unsigned char[:] buffer):
        super().__init__()
        self._next = None
        if buffer is None or buffer.size == 0:
            return
        _exc.check_error(x509.mbedtls_x509_crt_parse(
            &self._ctx, &buffer[0], buffer.size))

    def __cinit__(self):
        """Initialize a certificate (chain)."""
        x509.mbedtls_x509_crt_init(&self._ctx)

    def __dealloc__(self):
        """Unallocate all certificate data."""
        self.unset_next()
        x509.mbedtls_x509_crt_free(&self._ctx)

    def __next__(self):
        if self._next is None:
            raise StopIteration
        return self._next

    def __str__(self):
        cdef size_t osize = 2**24
        cdef char *output = <char *>malloc(osize * sizeof(char))
        cdef char *prefix = b""
        if not output:
            raise MemoryError()
        try:
            written = _exc.check_error(x509.mbedtls_x509_crt_info(
                &output[0], osize, prefix, &self._ctx))
            return output[:written].decode("utf8")
        finally:
            free(output)

    # RFC 5280, Section 4.1 Basic Certificate Fields
    # RFC 5280, Section 4.1.1 Certificate Fields

    @property
    def tbs_certificate(self):
        """The TBS (to be signed) certificate in DER format.

        See Also:
            RFC 5280, Section 4.1.1.1 TBS Certificate

        """
        return bytes(self._ctx.tbs.p[:self._ctx.tbs.len])

    @property
    def _signature_algorithm(self):
        """Cryptographic algorithm used by the CA to sign this CRT.

        See Also:
            RFC 5280, Section 4.1.1.2 Signature Algorithm

        """
        # The byte structure should be parsed according to RFC 5280.
        return bytes(self._ctx.sig_oid.p[:self._ctx.sig_oid.len])

    @property
    def signature_value(self):
        """Digital signature of the TBS certificate.

        See Also:
            RFC 5280, Section 4.1.1.3 Signature Value

        """
        return bytes(self._ctx.sig.p[:self._ctx.sig.len])

    # RFC 5280, Section 4.1.2 TBS Certificate

    @property
    def version(self):
        """The version of the encoded certificate.

        See Also:
            RF 5280, Section 4.1.2.1 Version

        """
        return int(self._ctx.version)

    @property
    def serial_number(self):
        """The certificate serial number.

        See Also:
            RFC 5280, Section 4.1.2.2 Serial Number

        """
        return int(_mpi.MPI.from_bytes(
            self._ctx.serial.p[:self._ctx.serial.len], "big"))

    # RFC 4.1.2.3 Signature
    @property
    def digestmod(self):
        return _new_hash(_md._digestmod(self._ctx.sig_md))

    @property
    def issuer(self):
        """Entity that has signed and issued the certificate.

        See Also:
            RFC 5280, Section 4.1.2.4 Issuer

        """
        cdef size_t osize = 200
        cdef char* c_buf = <char *>malloc(osize * sizeof(char))
        if not c_buf:
            raise MemoryError()
        try:
            written = x509.mbedtls_x509_dn_gets(
                &c_buf[0], osize, &self._ctx.issuer)
            return c_buf[:written].decode("utf8")
        finally:
            free(c_buf)

    @property
    def not_before(self):
        """Beginning of the validity of the certificate (inclusive).

        See Also:
            RFC 5280, Section 4.1.2.5 Validity

        """
        cdef x509.mbedtls_x509_time *valid_from = &self._ctx.valid_from
        return dt.datetime(
            valid_from[0].year, valid_from[0].mon, valid_from[0].day,
            valid_from[0].hour, valid_from[0].min, valid_from[0].sec)

    @property
    def not_after(self):
        """End of the validity of the certificate (inclusive).

        See Also:
            RFC 5280, Section 4.1.2.5 Validity

        """
        cdef x509.mbedtls_x509_time *valid_to = &self._ctx.valid_to
        return dt.datetime(
            valid_to[0].year, valid_to[0].mon, valid_to[0].day,
            valid_to[0].hour, valid_to[0].min, valid_to[0].sec)

    @property
    def subject(self):
        """Entity associated with the public key.

        See Also:
            RFC 5280, Section 4.1.2.6 Subject

        """
        cdef size_t osize = 200
        cdef char *c_buf = <char *>malloc(osize * sizeof(char))
        if not c_buf:
            raise MemoryError()
        try:
            written = x509.mbedtls_x509_dn_gets(
                &c_buf[0], osize, &self._ctx.subject)
            return c_buf[:written].decode("utf8")
        finally:
            free(c_buf)

    @property
    def subject_public_key(self):
        """The public key.

        See Also:
            RFC 5280, Section 4.1.2.7 Subject Public Key Info

        """
        cipher_type = _pk.CipherType(_pk.mbedtls_pk_get_type(&self._ctx.pk))
        cipher = {
            _pk.CipherType.RSA: _pk.RSA,
            _pk.CipherType.ECKEY: _pk.ECC,
        }.get(cipher_type, None)
        if cipher is None:
            raise NotImplementedError("unsupported cipher %r" % cipher_type)

        cdef size_t osize = _pk.PUB_DER_MAX_BYTES
        cdef unsigned char *c_buf = <unsigned char *>malloc(
            osize * sizeof(unsigned char))
        if not c_buf:
            raise MemoryError()
        try:
            ret = _exc.check_error(_pk.mbedtls_pk_write_pubkey_der(
                &self._ctx.pk, &c_buf[0], osize))
            return cipher.from_DER(c_buf[osize - ret:osize])
        finally:
            free(c_buf)

    # RFC 5280, Section 4.1.2.8 Unique Identifiers
    # RFC 5280, Section 4.1.2.9 Extensions
    # RFC 5280, Section 4.2 Certificate Extensions
    # RFC 5280, Section 4.2.1 Standard Extensions
    # RFC 5280, Section 4.2.1.1 Authority Key Identifier
    # RFC 5280, Section 4.2.1.2 Subject Key Identifier

    @property
    def key_usage(self):
        """Key usage extension (bitfield).

        See Also:
            RFC 5280, Section 4.2.1.3 Key Usage

        """
        return KeyUsage(self._ctx.key_usage)

    # RFC 5280, Section 4.2.1.4 Certificate Policies
    # RFC 5280, Section 4.2.1.5 Policy Mappings

    @property
    def subject_alternative_names(self):
        """Subject alternative name extension.

        See Also:
            RFC 5280, Section 4.2.1.6 Subject Alternative Name

        """
        cdef mbedtls_x509_sequence *item
        item = &self._ctx.subject_alt_names
        names = set()
        while item != NULL:
            names.add(item.buf.p[:item.buf.len].decode("utf8"))
            item = item.next
        return frozenset(names)

    # RFC 5280, Section 4.2.1.7 Issuer Alternative Name
    # RFC 5280, Section 4.2.1.8 Subject Directory Attributes

    @property
    def basic_constraints(self):
        """ca is true if the certified public key may be used
        to verify certificate signatures.

        See Also:
            - RFC 5280, Section 4.2.1.9 Basic Constraints
            - RFC 5280, `max_path_length`

        """
        max_path_length = int(self._ctx.max_pathlen)
        if max_path_length > 0:
            max_path_length -= 1
        ca = bool(self._ctx.ca_istrue)
        return BasicConstraints(ca, max_path_length)

    # RFC 5280, Section 4.2.1.10 Name Constraints
    # RFC 5280, Section 4.2.1.11 Policy Constraints
    # RFC 5280, Section 4.2.1.12 Extended Key Usage
    # RFC 5280, Section 4.2.1.13 CRL Distribution Points
    # RFC 5280, Section 4.2.1.14 Inhibit Any-Policy
    # RFC 5280, Section 4.2.1.15 Freshest CRL

    cdef set_next(self, CRT crt):
        self._next = crt
        self._ctx.next = &crt._ctx

    cdef unset_next(self):
        self._ctx.next = NULL
        self._next = None

    def check_revocation(self, CRL crl):
        """Return True if the certificate is revoked, False otherwise."""
        return bool(x509.mbedtls_x509_crt_is_revoked(&self._ctx, &crl._ctx))

    @classmethod
    def from_file(cls, path):
        path_ = str(path).encode("utf8")
        cdef const char* c_path = path_
        cdef CRT self = cls(None)
        _exc.check_error(x509.mbedtls_x509_crt_parse_file(&self._ctx, c_path))
        return self

    @classmethod
    def from_DER(cls, const unsigned char[:] buffer not None):
        if buffer.size == 0:
            raise ValueError("Cannot create %s from an empty buffer"
                             % cls.__name__)
        cdef CRT self = cls(None)
        _exc.check_error(x509.mbedtls_x509_crt_parse_der(
            &self._ctx, &buffer[0], buffer.size))
        return self

    @classmethod
    def from_PEM(cls, pem):
        return cls.from_DER(PEM_to_DER(pem))

    def to_DER(self):
        return bytes(self._ctx.raw.p[0:self._ctx.raw.len])

    def to_PEM(self):
        return DER_to_PEM(self.to_DER(), "Certificate")

    def sign(self, csr, issuer_key, not_before, not_after, serial_number,
             basic_constraints=BasicConstraints()):
        """Return a new, signed certificate for the CSR."""
        if not _pk.check_pair(
                self.subject_public_key,
                issuer_key):
            raise ValueError(
                "The issuer_key does not correspond to this certificate")
        cdef CRT crt = self.new(
            not_before=not_before,
            not_after=not_after,
            issuer=self.subject,
            issuer_key=issuer_key,
            subject=csr.subject,
            subject_key=csr.subject_public_key,
            serial_number=serial_number,
            digestmod=csr.digestmod,
            basic_constraints=basic_constraints)
        return crt

    @classmethod
    def selfsign(cls, csr, issuer_key, not_before, not_after, serial_number,
                 basic_constraints=BasicConstraints()):
        """Return a new, self-signed certificate for the CSR."""
        return cls.new(
            not_before=not_before,
            not_after=not_after,
            issuer=csr.subject,
            issuer_key=issuer_key,
            subject=csr.subject,
            subject_key=csr.subject_public_key,
            serial_number=serial_number,
            digestmod=csr.digestmod,
            basic_constraints=basic_constraints)

    def verify(self, crt):
        """Verify the certificate `crt`."""
        return self.subject_public_key.verify(
            crt.tbs_certificate, crt.signature_value, crt.digestmod.name)

    @staticmethod
    def new(not_before, not_after, issuer, issuer_key, subject, subject_key,
            serial_number, digestmod, basic_constraints=BasicConstraints()):
        """Return a new certificate."""
        return _CRTWriter(
            not_before, not_after, issuer, issuer_key,
            subject, subject_key, serial_number, digestmod,
            basic_constraints=basic_constraints).to_certificate()


cdef class _CRTWriter:
    """CRT writing context.

    This class should not be used directly.
    Use `CRT.new()` instead.

    """
    def __init__(self, not_before, not_after, issuer, issuer_key,
                 subject, subject_key, serial_number, digestmod,
                 basic_constraints=BasicConstraints()):
        super().__init__()
        self.set_validity(not_before, not_after)
        self.set_issuer(issuer)
        self.set_issuer_key(issuer_key)
        self.set_subject(subject)
        self.set_subject_key(subject_key)
        self.set_serial_number(serial_number)
        self.set_digestmod(digestmod)
        self.set_basic_constraints(basic_constraints)

    def __cinit__(self):
        """Initialize a CRT write context."""
        x509.mbedtls_x509write_crt_init(&self._ctx)

    def __dealloc__(self):
        """Free the contents of a CRT write context."""
        x509.mbedtls_x509write_crt_free(&self._ctx)

    def __getstate__(self):
        raise TypeError(f"cannot pickle {self.__class__.__name__!r} object")

    def to_DER(self):
        """Return the certificate in DER format.

        Warning:
            No RNG function is used.

        """
        cdef size_t osize = 4096
        cdef unsigned char *output = <unsigned char *>malloc(
            osize * sizeof(unsigned char))
        if not output:
            raise MemoryError()
        try:
            written = _exc.check_error(x509.mbedtls_x509write_crt_der(
                &self._ctx, &output[0], osize, NULL, NULL))
            return output[osize - written:osize]
        finally:
            free(output)

    def to_bytes(self):
        return self.to_DER()

    def to_certificate(self):
        """Return a Certificate object."""
        return CRT.from_DER(self.to_DER())

    def set_version(self, version=3):
        """Set the version for a certificate.

        Arg:
           version (int): The version between 1 and 3.

        """
        if version not in range(1, 4):
            raise ValueError("version not between 1 and 3")
        x509.mbedtls_x509write_crt_set_version(&self._ctx, version - 1)

    def set_serial_number(self, serial_number):
        """Set the serial number for a certificate.

        Arg:
            serial_number (int or bytes): The serial number.

        """
        if not serial_number:
            return
        cdef _mpi.MPI ser = _mpi.MPI(serial_number)
        x509.mbedtls_x509write_crt_set_serial(&self._ctx, &ser._ctx)

    def set_validity(self, not_before, not_after):
        """Set the validity period for a certificate.

        Args:
            not_before (datetime): Begin timestamp.
            not_after (datetime): End timestamp.

        """
        fmt = "%Y%m%d%H%M%S"
        # Keep reference to Python objects.
        not_before = not_before.strftime(fmt).encode("ascii")
        not_after = not_after.strftime(fmt).encode("ascii")
        cdef const char* c_not_before = not_before
        cdef const char* c_not_after = not_after
        _exc.check_error(x509.mbedtls_x509write_crt_set_validity(
            &self._ctx, c_not_before, c_not_after))

    def set_issuer(self, issuer):
        """Set the issuer name.

        Args:
            issuer (str): Comma-separated list of OID types and values:
                e.g. "C=UK,I=ARM,CN=mbed TLS CA"

        """
        # Keep reference to Python object.
        issuer_ = issuer.encode("utf8")
        cdef const char* c_issuer = issuer_
        _exc.check_error(x509.mbedtls_x509write_crt_set_issuer_name(
            &self._ctx, c_issuer))

    def set_subject(self, subject):
        """Set the subject name for a certificate.

        Args:
            subject (str): Subject name as a comma-separated list
                of OID types and values.
                e.g. "C=UK,O=ARM,CN=mbed TLS Server 1"

        """
        if subject is None:
            return
        # Keep reference to Python object.
        subject_ = subject.encode("utf8")
        cdef const char* c_subject = subject_
        _exc.check_error(x509.mbedtls_x509write_crt_set_subject_name(
            &self._ctx, c_subject))

    def set_digestmod(self, digestmod):
        """Set MD algorithm to use for the signature.

        Args:
            digestmod (MDBase): MD algorithm, for ex. `hash.sha1()`.

        """
        x509.mbedtls_x509write_crt_set_md_alg(&self._ctx, digestmod._type)

    def set_subject_key(self, _pk.CipherBase key):
        """Set the subject key.

        Args:
            key (CipherBase): Subject key.

        """
        x509.mbedtls_x509write_crt_set_subject_key(&self._ctx, &key._ctx)
        _exc.check_error(
            x509.mbedtls_x509write_crt_set_subject_key_identifier(&self._ctx))

    def set_issuer_key(self, _pk.CipherBase key):
        """Set the issuer key.

        Args:
            key (CipherBase): Issuer key.

        """
        x509.mbedtls_x509write_crt_set_issuer_key(
            &self._ctx, &key._ctx)
        _exc.check_error(
            x509.mbedtls_x509write_crt_set_authority_key_identifier(&self._ctx))

    def set_basic_constraints(self, basic_constraints):
        """Set the basic constraints extension for a CRT.

        Args:
            basic_constraints (BasicConstraints): The basic constraints.

        """
        if not basic_constraints:
            return
        _exc.check_error(x509.mbedtls_x509write_crt_set_basic_constraints(
            &self._ctx, int(basic_constraints[0]), basic_constraints[1]))


cdef class CSR(Certificate):
    """X.509 certificate signing request parser."""

    def __init__(self, const unsigned char[:] buffer):
        super().__init__()
        if buffer is None or buffer.size == 0:
            return
        _exc.check_error(x509.mbedtls_x509_csr_parse(
            &self._ctx, &buffer[0], buffer.size))

    def __cinit__(self):
        """Initialize a CSR."""
        x509.mbedtls_x509_csr_init(&self._ctx)

    def __dealloc__(self):
        """Unallocate all CSR data."""
        x509.mbedtls_x509_csr_free(&self._ctx)

    def __str__(self):
        cdef size_t osize = 2**24
        cdef char *output = <char *>malloc(osize * sizeof(char))
        cdef char *prefix = b""
        if not output:
            raise MemoryError()
        try:
            written = _exc.check_error(x509.mbedtls_x509_csr_info(
                &output[0], osize, prefix, &self._ctx))
            return output[:written].decode("utf8")
        finally:
            free(output)

    @property
    def digestmod(self):
        """Return the hash function used for the signature.

        See Also:
            RFC5280, Section 4.1.1.2 Signature Algorithm.

        """
        return _new_hash(_md._digestmod(self._ctx.sig_md))

    @property
    def version(self):
        """The version of the encoded certificate.

        See Also:
            RF 5280, Section 4.1.2.1 Version

        """
        return int(self._ctx.version)

    @property
    def subject(self):
        """Entity associated with the public key.

        See Also:
            RFC 5280, Section 4.1.2.6 Subject

        """
        cdef size_t osize = 200
        cdef char *c_buf = <char *>malloc(osize * sizeof(char))
        if not c_buf:
            raise MemoryError()
        try:
            written = x509.mbedtls_x509_dn_gets(
                &c_buf[0], osize, &self._ctx.subject)
            return c_buf[:written].decode("utf8")
        finally:
            free(c_buf)

    @property
    def subject_public_key(self):
        """The public key.

        See Also:
            RFC 5280, Section 4.1.2.7 Subject Public Key Info

        """
        cipher_type = _pk.CipherType(_pk.mbedtls_pk_get_type(&self._ctx.pk))
        cipher = {
            _pk.CipherType.RSA: _pk.RSA,
            _pk.CipherType.ECKEY: _pk.ECC,
        }.get(cipher_type, None)
        if cipher is None:
            raise ValueError("unsupported cipher %r" % cipher_type)

        cdef size_t osize = _pk.PUB_DER_MAX_BYTES
        cdef unsigned char *c_buf = <unsigned char *>malloc(
            osize * sizeof(unsigned char))
        if not c_buf:
            raise MemoryError()
        try:
            ret = _exc.check_error(_pk.mbedtls_pk_write_pubkey_der(
                &self._ctx.pk, &c_buf[0], osize))
            return cipher.from_DER(c_buf[osize - ret:osize])
        finally:
            free(c_buf)

    @classmethod
    def from_file(cls, path):
        path_ = str(path).encode("utf8")
        cdef const char* c_path = path_
        cdef CSR self = cls(None)
        _exc.check_error(x509.mbedtls_x509_csr_parse_file(&self._ctx, c_path))
        return self

    @classmethod
    def from_DER(cls, const unsigned char[:] buffer not None):
        if buffer.size == 0:
            raise ValueError("Cannot create %s from an empty buffer"
                             % cls.__name__)
        cdef CSR self = cls(None)
        _exc.check_error(x509.mbedtls_x509_csr_parse_der(
            &self._ctx, &buffer[0], buffer.size))
        return self

    @classmethod
    def from_PEM(cls, pem):
        return cls.from_DER(PEM_to_DER(pem))

    def to_DER(self):
        return bytes(self._ctx.raw.p[0:self._ctx.raw.len])

    def to_PEM(self):
        return DER_to_PEM(self.to_DER(), "Certificate Request")

    @staticmethod
    def new(subject_key, subject, digestmod):
        """Return a new CSR."""
        return _CSRWriter(subject_key, subject, digestmod).to_certificate()


cdef class _CSRWriter:
    """X.509 CSR writing context.

    This class should not be used directly.  Use `CSR.new()` instead.

    """
    def __init__(self, subject_key, subject, digestmod):
        super().__init__()
        self.set_subject_key(subject_key)
        self.set_subject(subject)
        self.set_digestmod(digestmod)

    def __cinit__(self):
        """Initialize a CSR context."""
        x509.mbedtls_x509write_csr_init(&self._ctx)

    def __dealloc__(self):
        """Free the contents of a CSR context."""
        x509.mbedtls_x509write_csr_free(&self._ctx)

    def __getstate__(self):
        raise TypeError(f"cannot pickle {self.__class__.__name__!r} object")

    def set_subject(self, subject):
        """Set the subject name for a CSR.

        Args:
            subject (str): Subject name as a comma-separated list
                of OID types and values.
                e.g. "C=UK,O=ARM,CN=mbed TLS Server 1"

        """
        if not subject:
            return
        # Keep reference to Python object.
        subject_ = subject.encode("utf8")
        cdef const char* c_subject = subject_
        _exc.check_error(x509.mbedtls_x509write_csr_set_subject_name(
            &self._ctx, c_subject))

    def set_subject_key(self, _pk.CipherBase key):
        """Set the key for the CSR.

        Args:
            key (CipherBase): Asymetric key to include.

        """
        x509.mbedtls_x509write_csr_set_key(&self._ctx, &key._ctx)

    def set_digestmod(self, digestmod):
        """Set MD algorithm to use for the signature.

        Args:
            digestmod (MDBase): MD algorithm, for ex. `hash.sha1()`.

        """
        x509.mbedtls_x509write_csr_set_md_alg(&self._ctx, digestmod._type)

    def to_DER(self):
        """Return the CSR in DER format.

        Warning:
            No RNG function is used.

        """
        cdef size_t osize = 4096
        cdef unsigned char *output = <unsigned char *>malloc(
            osize * sizeof(unsigned char))
        if not output:
            raise MemoryError()
        try:
            written = _exc.check_error(x509.mbedtls_x509write_csr_der(
                &self._ctx, &output[0], osize, NULL, NULL))
            return output[osize - written:osize]
        finally:
            free(output)

    def to_bytes(self):
        return self.to_DER()

    def to_certificate(self):
        """Return a CSR object."""
        return CSR.from_DER(self.to_DER())


class CRLEntry(namedtuple("CRLEntry", ["serial", "revocation_date"])):
    """An entry in the revocation list."""


cdef class CRL(Certificate):
    """X.509 revocation list."""

    def __init__(self, const unsigned char[:] buffer):
        super().__init__()
        self._next = None
        if buffer is None or buffer.size == 0:
            return
        _exc.check_error(x509.mbedtls_x509_crl_parse(
            &self._ctx, &buffer[0], buffer.size))

    def __cinit__(self):
        """Initialize a CRL (chain)."""
        x509.mbedtls_x509_crl_init(&self._ctx)

    def __dealloc__(self):
        """Unallocate all CRL data."""
        self.unset_next()
        x509.mbedtls_x509_crl_free(&self._ctx)

    def __next__(self):
        if self._next is None:
            raise StopIteration
        return self._next

    def __str__(self):
        cdef size_t osize = 2**24
        cdef char *output = <char *>malloc(osize * sizeof(char))
        cdef char *prefix = b""
        if not output:
            raise MemoryError()
        try:
            written = _exc.check_error(x509.mbedtls_x509_crl_info(
                &output[0], osize, prefix, &self._ctx))
            return output[:written].decode("utf8")
        finally:
            free(output)

    # RFC 5280, Section 5.1 CRL Fields
    # RFC 5280, Section 5.1.1 Certificate List Fields
    @property
    def tbs_certificate(self):
        """The TBS (to be signed) certificate in DER format.

        See Also:
            RFC 5280, Section 5.1.1.1 TBS Certificate

        """
        return bytes(self._ctx.tbs.p[:self._ctx.tbs.len])

    # RFC 5280, Section 5.1.1.2 Signature Algorithm

    @property
    def signature_value(self):
        """Cryptographic algorithm used by the CA to sign this CRL.

        See Also:
            RFC 5280, Section 5.1.1.3 Signature Algorithm

        """
        return bytes(self._ctx.sig.p[:self._ctx.sig.len])

    # RFC 5280, Section 5.1.2 Certificate List "To Be Signed"

    @property
    def version(self):
        """The version of the encoded certificate.

        See Also:
            RF 5280, Section 5.1.2.1 Version

        """
        return int(self._ctx.version)

    # RFC 5280, Section 5.1.2.2 Signature

    @property
    def issuer_name(self):
        """Entity that has signed and issued the certificate.

        See Also:
            RFC 5280, Section 5.1.2.3 Issuer Name

        """
        cdef size_t osize = 200
        cdef char* c_buf = <char *>malloc(osize * sizeof(char))
        if not c_buf:
            raise MemoryError()
        try:
            written = x509.mbedtls_x509_dn_gets(
                &c_buf[0], osize, &self._ctx.issuer)
            return c_buf[:written].decode("utf8")
        finally:
            free(c_buf)

    @property
    def this_update(self):
        """The issue date of this certificate.

        See Also:
            RFC 5280, Section 5.1.2.4 This Update

        """
        cdef x509.mbedtls_x509_time *this_update = &self._ctx.this_update
        return dt.datetime(
            this_update[0].year, this_update[0].mon, this_update[0].day,
            this_update[0].hour, this_update[0].min, this_update[0].sec)

    @property
    def next_update(self):
        """The date by which the next certificate will be issued.

        See Also:
            RFC 5280, Section 5.1.2.5 Next Update

        """
        cdef x509.mbedtls_x509_time *next_update = &self._ctx.next_update
        return dt.datetime(
            next_update[0].year, next_update[0].mon, next_update[0].day,
            next_update[0].hour, next_update[0].min, next_update[0].sec)

    @property
    def revoked_certificates(self):
        """The list of revoked certificates.

        See Also:
            RFC 5280, Section 5.1.2.6 Revoked Certificates

        """
        cdef x509.mbedtls_x509_crl_entry *item
        item = &self._ctx.entry
        revoked = []
        while item != NULL:
            revoked.append(CRLEntry(
                int(_mpi.MPI.from_bytes(
                    item.serial.p[:item.serial.len], "big")),
                dt.datetime(
                    item.revocation_date.year,
                    item.revocation_date.mon,
                    item.revocation_date.day,
                    item.revocation_date.hour,
                    item.revocation_date.min,
                    item.revocation_date.sec)))
            item = item.next
        return tuple(revoked)

    @classmethod
    def from_file(cls, path):
        path_ = str(path).encode("utf8")
        cdef const char* c_path = path_
        cdef CRL self = cls(None)
        _exc.check_error(x509.mbedtls_x509_crl_parse_file(&self._ctx, c_path))
        return self

    @classmethod
    def from_DER(cls, const unsigned char[:] buffer not None):
        if buffer.size == 0:
            raise ValueError("Cannot create %s from an empty buffer"
                             % cls.__name__)
        cdef CRL self = cls(None)
        _exc.check_error(x509.mbedtls_x509_crl_parse_der(
            &self._ctx, &buffer[0], buffer.size))
        return self

    @classmethod
    def from_PEM(cls, pem):
        return cls.from_DER(PEM_to_DER(pem))

    def to_DER(self):
        return bytes(self._ctx.raw.p[0:self._ctx.raw.len])

    def to_PEM(self):
        return DER_to_PEM(self.to_DER(), "X509 CRL")

    cdef set_next(self, CRL crl):
        self._next = crl
        self._ctx.next = &crl._ctx

    cdef unset_next(self):
        self._ctx.next = NULL
        self._next = None



# SPDX-License-Identifier: MIT
# Copyright (c) 2015, Elaborated Networks GmbH
# Copyright (c) 2018, Mathias Laurin

"""Generic message digest wrapper (hash algorithm)."""


cimport mbedtls._md as _md
from libc.stdlib cimport malloc, free

import binascii

import mbedtls.exceptions as _exc


__MD_NAME = (
    # Define as bytes to map to `const char*` without conversion.
    b"NONE",
    b"MD2",
    b"MD4",
    b"MD5",
    b"SHA1",
    b"SHA224",
    b"SHA256",
    b"SHA384",
    b"SHA512",
    b"RIPEMD160",
)


def __get_supported_mds():
    """Return the set of digests supported by the generic
    message digest module.

    See Also:
        mbedtls.tls.__get_supported_ciphersuites()

    """
    md_lookup = {n: v for n, v in enumerate(__MD_NAME)}
    cdef const int* md_types = _md.mbedtls_md_list()
    cdef size_t n = 0
    mds = set()
    while md_types[n]:
        mds.add(md_lookup[md_types[n]])
        n += 1
    return mds


def _digestmod(sig_md):
    return __MD_NAME[sig_md].decode("ascii").lower()


algorithms_guaranteed = ("md5", "sha1", "sha224", "sha256", "sha384", "sha512")
algorithms_available = {name.decode("ascii").lower()
                        for name in  __get_supported_mds()}


__all__ = algorithms_guaranteed + ("new", "algorithms_guaranteed",
                                   "algorithms_available")


cdef class MDBase:
    """Wrap and encapsulate the md library from mbed TLS.

    Parameters:
        name (str): The MD name known to mbed TLS.

    Attributes:
        digest_size (int): The size of the message digest, in bytes.
        block_size (int): The internal block size of the hash
            algorithm in bytes.
        name (str): The name of the message digest.

    """
    def __init__(self, name, _hmac):
        try:
            name_ = name.encode("ascii", "strict").upper()
        except (AttributeError, ValueError) as exc:
            # Not `str`-like.
            raise TypeError(name) from exc
        if not name in algorithms_available:
            raise ValueError("{} not available".format(name))
        cdef char *c_name = name_
        self._info = _md.mbedtls_md_info_from_string(c_name)
        _exc.check_error(_md.mbedtls_md_setup(&self._ctx, self._info, _hmac))

    def __cinit__(self):
        """Initialize an `md_context` (as NONE)."""
        _md.mbedtls_md_init(&self._ctx)

    def __dealloc__(self):
        """Free and clear the internal structures of ctx."""
        _md.mbedtls_md_free(&self._ctx)

    def __getstate__(self):
        raise TypeError(f"cannot pickle {self.__class__.__name__!r} object")

    def __str__(self):
        """Return the name of the message digest output."""
        return self.name

    @property
    def _type(self):
        """The type of the message digest output."""
        return _md.mbedtls_md_get_type(self._info)

    @property
    def digest_size(self):
        """The size of the resulting hash in bytes."""
        return _md.mbedtls_md_get_size(self._info)

    @property
    def block_size(self):
        """The internal block size of the hash algorithm in bytes."""
        return self._ctx.md_info.block_size

    @property
    def name(self):
        """The canonical name of the hashing algorithm."""
        return _md.mbedtls_md_get_name(self._info).decode("ascii").lower()

    cdef _finish(self, const unsigned char *output):
        return -0x5100  # Bad input data error.

    def update(self, const unsigned char[:] buffer not None):
        return -0x5100  # Bad input data error.

    def digest(self):
        """Return the digest output of `message`."""
        cdef size_t sz = self.digest_size
        cdef unsigned char* output = <unsigned char*>malloc(
            sz * sizeof(unsigned char))
        if not output:
            raise MemoryError()
        try:
            _exc.check_error(self._finish(output))
            return output[:self.digest_size]
        finally:
            free(output)

    def hexdigest(self):
        """Like digest except the digest is returned as a string object
        of double length.

        """
        return binascii.hexlify(self.digest()).decode("ascii")

    def copy(self):
        """Return a copy ("clone") of the MD object."""
        raise NotImplementedError


cdef class Hash(_md.MDBase):

    """Wrap and encapsulate hash calculations.

    This class is a wrapper for the hash calculations in the md module
    of mbed TLS.  The interface follows the recommendation from PEP 452
    for unkeyed hashes.

    Parameters:
        name (str): The MD name known to mbed TLS.

    Attributes:
        digest_size (int): The size of the message digest, in bytes.
        block_size (int): The internal block size of the hash
            algorithm in bytes.
        name (str): The name of the message digest.

    """
    def __init__(self, name, buffer=None):
        super().__init__(name, 0)
        _exc.check_error(_md.mbedtls_md_starts(&self._ctx))
        self.update(buffer)

    def update(self, const unsigned char[:] buffer):
        """Update the hash object with the `buffer`."""
        if buffer is None or buffer.size == 0:
            return
        _exc.check_error(
            _md.mbedtls_md_update(&self._ctx, &buffer[0], buffer.size))

    cdef _finish(self, unsigned char *output):
        """Return the digest output of `message`."""
        return _md.mbedtls_md_finish(&self._ctx, output)

    def copy(self):
        """Return a copy ("clone") of the hash object."""
        obj = Hash(self.name)
        _exc.check_error(_md.mbedtls_md_clone(&obj._ctx, &self._ctx))
        return obj


cdef class Hmac(_md.MDBase):

    """Wrap and encapsulate HMAC calculations.

    This class is a wrapper for the HMAC calculations in the md module
    of mbed TLS.  The interface follows the recommendation from PEP 452
    for keyed hashes.

    Parameters:
        key (bytes): The key to use.
        name (str): The MD name known to mbed TLS.

    Warning:
        The message is cleared after calculation of the digest.  Only
        call :meth:`digest` or :meth:`hexdigest` once per message.

    Attributes:
        digest_size (int): The size of the message digest, in bytes.
        block_size (int): The internal block size of the hash
            algorithm in bytes.
        name (str): The name of the message digest.

    """
    def __init__(
        self, const unsigned char[:] key not None, name, buffer=None
    ):
        super().__init__(name, 1)
        if not key.size:
            key = b"\0"
        _exc.check_error(_md.mbedtls_md_hmac_starts(&self._ctx, &key[0], key.size))
        self.update(buffer)

    def update(self, const unsigned char[:] buffer):
        """Update the HMAC object with `buffer`."""
        if buffer is None or buffer.size == 0:
            return
        _exc.check_error(
            _md.mbedtls_md_hmac_update(&self._ctx, &buffer[0], buffer.size))

    cdef _finish(self, unsigned char *output):
        """Return the HMAC of key and message."""
        ret = _md.mbedtls_md_hmac_finish(&self._ctx, output)
        if ret != 0:
            return ret
        return _md.mbedtls_md_hmac_reset(&self._ctx)

    def copy(self):
        """Return a copy ("clone") of the HMAC object.

        Warning:
            Not implemented in mbed TLS, raises NotImplementedError.

        """
        raise NotImplementedError



# SPDX-License-Identifier: MIT
# Copyright (c) 2016, Elaborated Networks GmbH
# Copyright (c) 2019, Mathias Laurin

"""Data Encryption Standard (DES) cipher developed by IBM
in the 70's."""


from mbedtls.exceptions import TLSError

from . import _cipher

__all__ = ["block_size", "key_size", "new"]


block_size = 8
key_size = 8


def new(key, mode, iv=None):
    """Return a `Cipher` object that can perform DES encryption and
    decryption.

    Data Encryption Standard (DES) cipher developed by IBM in the 70's.

    Parameters:
        key (bytes or None): The key to encrypt decrypt.  If None,
            encryption and decryption are unavailable.
        mode (Mode): The mode of operation of the cipher.
        iv (bytes or None): The initialization vector (IV).  The IV is
            required for every mode but ECB and CTR where it is ignored.
            If not set, the IV is initialized to all 0, which should not
            be used for encryption.

    """
    mode = _cipher.Mode(mode)
    if len(key) != key_size:
        raise TLSError(msg="key size must be 16 bytes, got %r" % len(key))
    if mode not in {_cipher.Mode.ECB, _cipher.Mode.CBC}:
        raise TLSError(msg="unsupported mode %r" % mode)
    name = ("DES-%s" % mode.name).encode("ascii")
    return _cipher.Cipher(name, key, mode, iv)
